#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  2 15:35:55 2024

@author: roncofaber
"""

import numpy as np
from sklearn.decomposition import PCA
import ase
import ase.visualize

#%%
class PCAmodel():
    
    def __init__(self, clusters):
        
        # get positions
        X = np.array([clu.get_positions().flatten() for clu in clusters])

        # get mean (meaningful because of SU)
        mean_pos = np.mean(X, axis=0).reshape(-1,3)
        
        # save mean structure
        self.mean_structure = clusters[0].copy()
        self.mean_structure.set_positions(mean_pos)
        
        # perform mass weighting
        self.mass_weight = np.sqrt(np.repeat(clusters[0].get_masses(),3))
        X_w = self.mass_weight*X

        # FIT PCA MODEL
        pca_model = PCA()
        pca_model.fit(X_w)

        # save model
        self.pca_model = pca_model
        self.pca_modes = self.generate_pca_modes(pca_model)

        return
    
    def generate_pca_modes(self, pca_model):
        
        # x_pca = np.dot(X_w, evecs) #TODO understand if this is needed
        # mode_scale = np.sqrt(np.sum(np.var(np.dot(x_pca, evecs[:, 0]))))
        
        evals = pca_model.singular_values_
        evecs = pca_model.components_
        
        pca_modes = []
        for pca_comp in range(pca_model.n_components_):

            displacement = np.dot(evecs[pca_comp], np.sqrt(evals[pca_comp]))
            displacement = displacement/self.mass_weight
            displacement = displacement.reshape(-1, 3)

            fac = np.linspace(0,1,50)
            tclus = []
            for ii in fac:
                tclu = self.mean_structure.copy()
                tclu.set_pbc(False)
                tclu.set_positions(self.mean_structure.get_positions() + ii*displacement)
                tclus.append(tclu)

            pca_modes.append(tclus)
        
        return pca_modes
    
    def transform(self, cluster):
        
        if not isinstance(cluster, list):
            cluster = [cluster]
            
        X = np.array([clu.get_positions().flatten() for clu in cluster])
        X_w = self.mass_weight*X
        
        return self.pca_model.transform(X_w)
    
    def view(self, pca_idx):
        
        ase.visualize.view(self.pca_modes[pca_idx])
        
        return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions to read basis set and get info

Created on Thu Jul 13 15:57:59 2023

@author: roncoroni
"""

import numpy as np

#%%

# function that finds the permutation needed to go from Qchem internal order
# to format of fchk files (IQmol compatible)
def calculate_shell_reordering(shell, start_idx):

    index_list = []

    if shell["shell_type"].startswith("d"):
        index_list = [ii + start_idx for ii in [4,2,0,1,3]]
    elif shell["shell_type"].startswith("f"):
        index_list = [ii + start_idx for ii in [6,4,2,0,1,3,5]]
    elif any([shell["shell_type"].startswith(ss) for ss in ["s", "p"]]):
        index_list = [ii + start_idx for ii in range(shell["functions"])]
    else:
        raise Exception("Sorry, shell type {} not recognized".format(
            shell["shell_type"]))

    return index_list


# get info from basis set
def get_basis_set_info(basis):
    atom_coeffs = []
    atom_labels = []
    symbols     = []
    nbasis = 0
    indexing    = []
    for cc, atom in enumerate(basis['atoms']):
        istart = nbasis
        atom_label = []
        for shell in atom['shells']:

            # store internal ordering
            indexing.extend(calculate_shell_reordering(shell, nbasis))

            # update number of basis
            nbasis += shell['functions']

            # store shell types
            atom_label.extend(shell['functions']*[shell['shell_type']])

        atom_coeffs.append(np.array(range(istart, nbasis)))
        atom_labels.append(np.array(atom_label))
        symbols.append(atom["symbol"])

    return atom_coeffs, atom_labels, symbols, nbasis, indexing

# calculate AO permutation from structure permutation
def get_AO_permutation(mol, permutation):
    
    # permute AO according to the permutation
    aoslices = mol.aoslice_by_atom()
    
    AO_permutation = []
    for idx in permutation:
        tslice = list(range(aoslices[idx][2], aoslices[idx][3]))
        
        AO_permutation.append(tslice)
    
    return np.concatenate(AO_permutation)

def get_l_val(mol):
    
    l_values = {
        "s" : 0,
        "p" : 1,
        "d" : 2,
        "f" : 3,
        "g" : 4,
        }
    return np.array([l_values[ii[2][-1]] for ii in mol.ao_labels(fmt=False)])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 16:56:01 2023

@author: roncoroni
"""

import numpy as np

#%%

# Function to run MBXAS of pyscf calculators
def run_MBXAS_pyscf(mol, gs_calc, fch_calc, gs_orb_idx, channel=1, xch_calc=None):

    # Calculate dipole integrals and basis set overlap matrix
    dipole = mol.intor('int1e_r')  # Dipole moment integrals
    basis_overlap = mol.intor("int1e_ovlp")  # Basis function overlap matrix

    # Calculate FCH dipole matrix 
    dipole_KS = fch_calc.mo_coeff[channel].T @ dipole @ fch_calc.mo_coeff[channel]

    # Calculate many-body overlap matrix
    mb_overlap = fch_calc.mo_coeff[channel].T @ basis_overlap @ gs_calc.mo_coeff[channel]

    # Index of the excited orbital in FCH calculation
    exc_orb_idx = np.where(fch_calc.mo_occ[channel] == 0)[0][0] 

    # Occupied and unoccupied orbital indices for GS and FCH
    occ_idxs_fch = np.where(fch_calc.mo_occ[channel] == 1)[0]
    occ_idxs_gs = np.delete(np.where(gs_calc.mo_occ[channel] == 1)[0], gs_orb_idx)
    uno_idxs_fch = np.where(fch_calc.mo_occ[channel] == 0)[0][1:] 

    # Extract occupied block of the MB matrix
    AMat = mb_overlap[np.ix_(occ_idxs_fch, occ_idxs_gs)] 

    # Determinant of AMat
    ADet = np.linalg.det(AMat)

    # Extract unoccupied block of the MB matrix
    APrimeMat = mb_overlap[np.ix_(uno_idxs_fch, occ_idxs_gs)]

    # Calculate KMat 
    KMat = APrimeMat @ np.linalg.inv(AMat)

    # Transition dipole moments from excited orbital
    chb_xmat = dipole_KS[:, :, exc_orb_idx]
    chb_xmat_occ = chb_xmat[:,occ_idxs_fch]
    chb_xmat_uno = chb_xmat[:,uno_idxs_fch]

    # Calculate absorption spectrum
    absorption = ADet*(chb_xmat_uno - (KMat @ chb_xmat_occ.T).T)

    # Get excitation energies
    energies = fch_calc.mo_energy[channel][uno_idxs_fch]

    # Optional: Correct energies with exchange-correlation calculation
    if xch_calc is not None:
        energies += xch_calc.e_tot - gs_calc.e_tot - np.min(energies)

    # Return results 
    return energies, absorption, mb_overlap, dipole_KS, basis_overlap 

# # Function to run MBXAS of pyscf calculators
# def run_MBXAS_pyscf(mol, gs_calc, fch_calc, gs_orb_idx, channel=1, xch_calc=None):
    
#     # calculate dipole and basis overlap
#     dipole = mol.intor('int1e_r')
#     basis_overlap = mol.intor("int1e_ovlp")
    
#     # calculate FCH dipole matrix
#     dipole_KS = fch_calc.mo_coeff[channel].T@dipole@fch_calc.mo_coeff[channel]
    
#     # calculate many body overlap
#     mb_overlap = fch_calc.mo_coeff[channel].T@basis_overlap@gs_calc.mo_coeff[channel]
    
#     # index of excited oribtal (first 0 element)
#     exc_orb_idx = np.where(fch_calc.mo_occ[channel] == 0)[0][0]
    
#     # define indexes of occupied and unoccupied orbitals for GS and FCH
#     occ_idxs_fch = np.where(fch_calc.mo_occ[channel] == 1)[0]
#     occ_idxs_gs  = np.delete(np.where(gs_calc.mo_occ[channel] == 1)[0], gs_orb_idx)
#     uno_idxs_fch = np.where(fch_calc.mo_occ[channel] == 0)[0][1:] # remove first unocc as it's the CH
    
#     # extract from the MB matrix only what is occupied
#     AMat = mb_overlap[np.ix_(occ_idxs_fch, occ_idxs_gs)] # np.ix_ works with rows and cols
    
#     # calculate determinant of AMat
#     ADet = np.linalg.det(AMat)
    
#     # extract A': size n_unocc, n_occ
#     APrimeMat = mb_overlap[np.ix_(uno_idxs_fch, occ_idxs_gs)]
    
#     # calculate KMat from A' and A
#     KMat = APrimeMat @ np.linalg.inv(AMat)
    
#     # get transition contribution from the excited orbital
#     chb_xmat = dipole_KS[:, :, exc_orb_idx]
#     chb_xmat_occ = chb_xmat[:,occ_idxs_fch]
#     chb_xmat_uno = chb_xmat[:,uno_idxs_fch]
    
#     # calculate absorption (FCH contribution - CH contribution)
#     absorption = ADet*(chb_xmat_uno - (KMat @ chb_xmat_occ.T).T)
    
#     # get energies
#     energies = fch_calc.mo_energy[channel][uno_idxs_fch]

#     if xch_calc is not None:
#         energies += xch_calc.e_tot - gs_calc.e_tot - np.min(energies)

#     return energies, absorption, mb_overlap, dipole_KS, basis_overlap


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 23 10:32:15 2023

@author: roncofaber
"""

import os
from pyqchem.parsers.basic import basic_parser_qchem
from pyqchem.parsers.parser_optimization import basic_optimization
#%%

#Read outputs
def read_qchem_job(output, run_path, jobtype):
    
    if not os.path.isdir(run_path):
        print("Doesnt exist!")
    
    file_to_read = "{}/{}".format(run_path, output)
    
    # print(file_to_read)
    with open(file_to_read, "r") as fin:
        output = fin.read()
        
    if jobtype == "sp":
        parsed_data = basic_parser_qchem(output)
        
    elif jobtype == "opt":
        parsed_data = basic_optimization(output)
        
    return output, parsed_data


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 17:41:10 2023

@author: roncoroni
"""

import sys
from io import StringIO

import logging
# set up logger object
logging.basicConfig(
    level   = logging.INFO,
    format  = "%(asctime)s |%(message)s",  # Modified format
    datefmt = "[%H:%M:%S]",
)

#%%

def configure_logger(level):
    
    # Define a mapping from user input to logging levels
    level_mapping = {
        1: logging.ERROR,
        2: logging.WARNING,
        3: logging.INFO,
        4: logging.DEBUG,
        5: logging.CRITICAL,
    }
    
    # level
    level = level_mapping[level]
    
    # Get the root logger
    root_logger = logging.getLogger()
    
    # Clear existing handlers
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)
    
    # Set the logging level
    root_logger.setLevel(level)
    
    # Create a console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    
    # Create a formatter
    formatter = logging.Formatter("%(asctime)s |%(message)s", datefmt = "[%H:%M:%S]")
    
    # Set the formatter for the handler
    console_handler.setFormatter(formatter)
    
    # Add the handler to the root logger
    root_logger.addHandler(console_handler)
    
    return




# Logger to both print to terminal but store the output as string
class Logger(object):
    def __init__(self, print_to_terminal=True):
        self.print_to_terminal = print_to_terminal
        if print_to_terminal:
            self.terminal_write = sys.stdout.write
        else:
            self.terminal_write = lambda message: None  # A dummy function that does nothing
        self.log = StringIO()
        
    def write(self, message):
        self.terminal_write(message)
        self.log.write(message)

    def flush(self):
        # this flush method is needed for python 3 compatibility.
        # this handles the flush command by doing nothing.
        # you might want to specify some extra behavior here.
        pass
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 14 15:04:46 2023

@author: roncoroni
"""

import numpy as np

#%%


# return copy of input as list if not one
def as_list(inp):
    return [inp] if not isinstance(inp, list) else inp.copy()

# return list of indexes from mixed input of indexes and string (elements)
def atoms_to_indexes(system, symbols):

    # check if symbols is a list of strings
    if symbols == "all":
        return list(range(len(system.get_chemical_symbols())))

    symbols = as_list(symbols)

    indexes = []
    for symbol in symbols:
        if not isinstance(symbol, str):
            indexes.append(symbol)
        else:
            for cc, atom in enumerate(system.get_chemical_symbols()):
                if atom == symbol:
                    indexes.append(cc)
    return np.unique(indexes).tolist()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 26 10:33:37 2023

@author: roncofaber
"""

# data manipulation
import numpy as np
from functools import reduce
import dill

# pymbxas utils
from pymbxas.build.structure import rotate_structure, ase_to_mole
from pymbxas.utils.basis import get_AO_permutation, get_l_val
from pymbxas.mbxas.broaden import get_mbxas_spectra
from pymbxas.io.write import write_data_to_fchk

# can we use the sea urchin here?
try:
    import sea_urchin.alignement.align as ali
    has_SU = True
except:
    has_SU = False

# pyscf stuff
from pyscf import lo
from pyscf.lo import iao, orth

#%%

class Spectra():

    def __init__(self,
                 pyscf_obj,
                 excitation=None,
                 ):
        
        # read spectra from pyscf object
        if "calculators.pyscf" in str(type(pyscf_obj)):
            self.__initialize_spectra(pyscf_obj, excitation)
        else:
            self.__restart(pyscf_obj)
        
        return

    # function that reads and initialize the spectra object #TODO this is mostly to be updated
    def __initialize_spectra(self, pyscf_obj, excitation):
        
        if excitation is None:
            assert len(pyscf_obj.excitations) == 1, "Specify one excitation"
            excitation = 0
                
        
        # retreive calculation details
        self.mol       = pyscf_obj.mol
        self.structure = pyscf_obj.structure
        
        self.calc_settings = pyscf_obj.parameters
        
        # get excitation data
        data  = pyscf_obj.excitations[excitation].data["fch"]
        mbxas = pyscf_obj.excitations[excitation].mbxas
        channel = pyscf_obj.excitations[excitation].channel
        
        self.gs_energy = pyscf_obj.gs_data.e_tot
        self.energies  = mbxas["energies"]
        self.amplitude = mbxas["absorption"]
        
        self._mo_coeff = data.mo_coeff[channel]
        self._mo_occ   = data.mo_occ[channel]
        
        # metadata
        self._el_labels = [-1]*self.CMO.shape[1]
        self._label     = -1
        
        return
    
    def __restart(self, pyscf_obj):
        
        #load a pkl or dict
        if isinstance(pyscf_obj, dict):
            self.__dict__ = pyscf_obj.copy()
        elif isinstance(pyscf_obj, str):
            self.__dict__ = self.__pkl_to_dict(pyscf_obj)
            
        self.make_mol()
            
        return
    
    def __pkl_to_dict(self, filename):
        with open(filename, 'rb') as fin:
            data = dill.load(fin)
        return data
    
    def transform(self, rot=None, tr=None, perm=None,
                       inv=None, atype=None):
        
        if rot is None:
            rot = np.eye(3)
        if tr is None:
            tr = np.zeros(3)
        # no inversion? Use det of rot #TODO
        if inv is None:
            inv = np.round(np.linalg.det(rot))
        if perm is None:
            perm = list(range(len(self.structure)))
        if atype is None:
            atype = "fastoverlap"
            print("Assuming FO as type")
        
        # generate rotated structure
        structure = rotate_structure(self.structure, rot, tr, perm, inv, atype)
        
        # convert to mole
        mol = ase_to_mole(structure, verbose=0, **self.calc_settings)
                        
        # generate rotation matrix from rotM
        U = mol.ao_rotation_matrix(rot)
        
        # get permutation of AOs to match structure perm
        AO_permutation = get_AO_permutation(mol, perm)
        
        # calculate inversion contribution
        inv_A = inv**get_l_val(mol)
        
        # calculate rotated MOs
        ali_MOs = (inv_A*U).T.dot(self._mo_coeff[AO_permutation])
        
        # reassign variables
        self.structure = structure
        self._mo_coeff = ali_MOs
        self.mol       = mol
        self.amplitude = inv*rot@(self.amplitude)
        
        return
    
    def align_to_reference(self, reference, alignment, subset=None):
        
        assert has_SU, "Please, install also Sea Urchin to do this"
        
        if subset is None:
            structure = self.structure
        else:
            structure = self.structure[subset]
        
        rot, tr, perm, inv, _ = ali.get_RTPI(structure, reference, alignment)
        
        assert inv in [-1, 1]
        
        self.transform(rot=rot, tr=tr, perm=perm, inv=inv, atype=alignment["type"])
        
        return
    
    def get_CMO_projection(self, AO_to_proj, mol=None):
        
        if mol is None:
            mol = self.mol
        
        basis_ovlp = mol.intor_symmetric("int1e_ovlp")
        
        return (self.CMO.T@basis_ovlp@AO_to_proj).T
    
    def get_CMO_orth_proj(self, orth_method="meta-lowdin"):
        
        # calculate basis overlap
        basis_ovlp = self.mol.intor_symmetric('int1e_ovlp')
        
        # get orth AOs
        lao = lo.orth.orth_ao(self.mol, orth_method, s=basis_ovlp)
        
        # return CMO proj on LAOs
        return (self.CMO.T@basis_ovlp@lao).T
    
    
    # generate iaos given a structure and a basis (assumes FCH)
    def make_iaos(self, minao="minao"):
        
        maxidx = np.where(self._mo_occ == 1)[0].max()
        
        b_ovlp = self.mol.intor_symmetric('int1e_ovlp')
        
        iaos = iao.iao(self.mol, self._mo_coeff[:,:maxidx], minao=minao)
        
        return np.dot(iaos, orth.lowdin(reduce(np.dot, (iaos.T,b_ovlp,iaos))))
    
    def get_mbxas_spectra(self, axis=None, sigma=0.005, npoints=3001, tol=0.01,
                          erange=None, el_label=None):
        
        if el_label is not None:
            assert self._el_labels is not None, "first run el. clustering"
            
            idxs = self._el_labels == el_label
            
            amplitude = self.amplitude[:,idxs]
            energies  = self.energies[idxs] 
        else:
            amplitude = self.amplitude
            energies  = self.energies
        
        erange, spectras = get_mbxas_spectra(energies, amplitude,
                                              sigma=sigma, npoints=npoints,
                                              tol=tol, erange=erange)

        if axis is None:
            spectras = np.mean(spectras, axis=0)
        else:
            spectras = spectras[axis]
        
        return erange, spectras
        
    def get_amplitude_tensor(self):
        return np.einsum("ij,jk->ikj", self.amplitude, self.amplitude.T)
    
    # get CMOs
    @property
    def CMO(self): 
        uno_idxs = np.where(self._mo_occ == 0)[0][1:]
        return self._mo_coeff[:, uno_idxs]
    
    @property
    def label(self):
        return self._label
    
    
    def write_CMO2fchk(self, center=True, oname="spectra_CMO.fchk",
                       mo_coeff=None):
        
        if mo_coeff is None:
            mo_coeff = self.CMO
        
        write_data_to_fchk(self.mol,
                           mo_coeff  = mo_coeff,
                           mo_energy = self.energies,
                           mo_occ    = np.zeros((2,len(self.energies))),
                           center    = center,
                           oname     = oname)
        
        return
    
    # use this to remake mol obj (useful for reloading with dill)
    def make_mol(self):
        self.mol = ase_to_mole(self.structure, verbose=0, **self.calc_settings)
        return
    
    def _prepare_for_save(self):
        """Subclasses can override this to customize saving behavior.

        Returns:
            dict: A dictionary containing data to be serialized.
        """
        data = self.__dict__.copy()
        del data["mol"]
        return data
    
    def save(self, filename="spectra.pkl"):
        """Saves the object to a file."""
        
        data = self._prepare_for_save()
      
        # Save to file
        with open(filename, 'wb') as fout:
            dill.dump(data, fout)
            
        return
    
    # use it to return a copy of the spectra
    def copy(self):
        data = self._prepare_for_save()
        return Spectra(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 26 16:59:25 2023

@author: roncofaber
"""

import argparse
import os, sys


from pymbxas.io import write, run
from pymbxas import __mbxasdir__
#%%

def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

#%%
    
parser = argparse.ArgumentParser()

parser.add_argument("-np", "--npoints", dest="gridP", nargs='?',
                    help="Number of grid points.",
                    default=301, type=int)

parser.add_argument("-he", "--highE", dest="highE", nargs='?',
                    help="High energy limit.",
                    default=0.0, type=float)

parser.add_argument("-le", "--lowE", dest="lowE", nargs='?',
                    help="Low energy limit.",
                    default=0.0, type=float)

parser.add_argument("-s", "--sigma", dest="sigma", nargs='?',
                    help="Gaussian broadening RIXS",
                    default=0.3, type=float)

parser.add_argument("-r", "--rixs", dest="DoRIXS", nargs='?',
                    help="Do RIXS.",
                    default=False, type=str2bool, const=True)

parser.add_argument("-a", "--align", dest="do_align", nargs='?',
                    help="Do alignment.",
                    default=False, type=str2bool, const=True)


def main(argv=sys.argv[1:]):
    
    # initialize variables
    run_path = os.getcwd()
    
    # default params
    mbxas_params = {
        "gridP"    : 100,
        "highE"    : 0,
        "lowE"     : 0,
        "sigma"    : 0.3,
        "do_align" : False,
        "DoRIXS"   : False,

        "Gamma"         : "0+0.3j",
        "check_amp"     : True,
        "printsticks"   : True,
        "printspec"     : True,
        "Dodebug"       : False,
        "calc_abs"      : True,
        "printinteg"    : False,
        "input_file"    : "{}/qchem.input".format(run_path),
        "output_file"   : "{}/qchem.output".format(run_path),
        "printanalysis" : True
    }
    
    # parse arguments
    attributes = parser.parse_args(argv)
    
    # update default parameters
    mbxas_params.update(attributes.__dict__)
    
    # write input file
    write.write_mbxas_input(mbxas_params, run_path)

    # run MBXAS job
    run.submit_mbxas_job(run_path, __mbxasdir__)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 17:03:09 2023

@author: roncoroni
"""

import numpy as np
from ase import units
Ha = units.Ha
#%%

def gaussian_broadening(x, sigma):
    return np.exp(-x**2 / (2 * sigma**2)) / (sigma * np.sqrt(2 * np.pi))


def broadened_spectrum(x, energies, intensities, sigma):

    x_shifted = x[:, np.newaxis] - energies  # Efficiently subtract energies from each x value

    return np.sum(intensities * gaussian_broadening(x_shifted, sigma)[:, np.newaxis], axis=2)


# function to get MBXAS spectra
def get_mbxas_spectra(energies, intensities, sigma=0.005, npoints=3001, tol=0.01,
                      erange=None, isotropic=False):
    
    if erange is not None:
        min_E, max_E = np.array(erange)/Ha
        tol = 0
    else:
        min_E = np.min(energies)
        max_E = np.max(energies)

    dE = max_E - min_E
    
    energy = np.linspace(min_E - tol*dE, max_E + tol*dE, npoints)
    rel_idxs = (energies > min_E-dE) & (energies < max_E+dE)
    
    if isotropic:
        spectras = broadened_spectrum(energy, energies[rel_idxs],
                                         intensities[rel_idxs], sigma)  # Vectorized calculation
        
    # else:
    if not isotropic:
        spectras = broadened_spectrum(energy, energies[rel_idxs],
                                     intensities[:,rel_idxs]**2,
                                     sigma)  # Vectorized calculation
    
    return Ha*energy, spectras.T



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 29 10:08:26 2023

@author: roncoroni
"""
import copy
import collections

from pyqchem import Structure, QchemInput
from pyqchem.qc_input import CustomSection
from pymbxas.utils.check_keywords import determine_occupation

#%%

gs_def_params = {
    "extra_rem_keywords" : {"TRANS_MOM_SAVE" : True,
                            # "BOYSCALC"       : "2",
                            },
    }

xas_def_params = {
    "extra_rem_keywords" : {
        "TRANS_MOM_READ" : True,
        'MOM_METHOD'     : 'MOM', # or IMOM?
        "SCF_GUESS"      : 'read',
        "MOM_START"      : 1 ,
        "use_libqints"   : 1
    },
    }

def update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def update_input(qchem_params, extra_input, calc_type, use_boys=False,
                 scf_guess=None, occupation=None):

    # copy input
    qchem_params = copy.deepcopy(qchem_params)
    extra_input  = copy.deepcopy(extra_input)

    # remove mom_type #TODO ugly and temporary
    if "mom_type" in qchem_params:

        mom_type = qchem_params["mom_type"]
        del qchem_params["mom_type"]

        if calc_type == "gs":
            pass
        else:
            if mom_type == 0: # 0 --> all MOM
                extra_input["extra_rem_keywords"]["MOM_METHOD"] = "MOM"
            if mom_type == 1: # 1 --> all IMOM
                extra_input["extra_rem_keywords"]["MOM_METHOD"] = "IMOM"

    # update with default params
    qchem_params = update(qchem_params, extra_input)

    if use_boys:
        qchem_params["extra_rem_keywords"]["BOYSCALC"] = 2
    else:
        qchem_params["extra_rem_keywords"]["BOYSCALC"] = 0

    if scf_guess is not None:
        qchem_params["scf_guess"] = scf_guess

    # add extra sections if nonexistend #TODO can be cleaner code
    if "extra_sections" not in qchem_params:
        qchem_params["extra_sections"] = None

    # check occupation if needed
    if occupation is not None:
        # check occupation format
        occ_section = CustomSection(title='occupied',
                                    keywords={' ' : determine_occupation(occupation)})

        if isinstance(qchem_params["extra_sections"], list):
            qchem_params["extra_sections"].append(occ_section)
        else:
            qchem_params["extra_sections"] = occ_section

    return qchem_params

# Function to generate an input for a QCHEM calculation,
# specify calc type to add default params needed for MBXAS
def make_qchem_input(molecule, charge, multiplicity,
                     qchem_params, calc_type, occupation = None,
                     scf_guess = None, use_boys=False):

    # make molecule
    molecule_str = Structure(
        coordinates  = molecule.get_positions(),
        symbols      = molecule.get_chemical_symbols(),
        charge       = charge,
        multiplicity = multiplicity)

    # update params depending on calculation
    if calc_type == "gs":
        qchem_params = update_input(qchem_params, gs_def_params, calc_type,
                                    use_boys, scf_guess, occupation)
    elif calc_type in ["fch", "xch"]:
        qchem_params = update_input(qchem_params, xas_def_params, calc_type,
                                    use_boys, scf_guess, occupation)

    # generate input
    qchem_input = QchemInput(
            molecule_str,
            **qchem_params,
            )

    return qchem_input


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Series of functions to help fixing keywords of PyMBXAS dict.

Created on Mon Jun 26 11:44:43 2023

@author: roncofaber
"""

#%%

# default excitation dict
def_exc_dict = {
    "channel" : "beta",
    "eject"   : False,
    "inject"  : False,
    }

# Function to transform an occ. dict to input ready for QCHEM
def determine_occupation(occupation):

    if isinstance(occupation, str):
        return occupation

    elif isinstance(occupation, dict):

        # check inject
        if "inject" not in occupation:
            occupation["inject"] = False

        # generate occupation string
        occ_string = ""
        for cc, channel in enumerate(["alpha", "beta"]):

            indexes = [str(x) for x in range(1, occupation["nelectrons"][cc]+1)
                       if not (x == occupation["eject"]
                               and channel == occupation["channel"])]

            if occupation["inject"] and channel == occupation["channel"]:
                indexes.append(str(occupation["inject"]))

            occ_string += " ".join(indexes)

            occ_string += "\n  "

        return occ_string
    
#TODO expand this to work with mixed PBCs, works only for full pbc at the moment
def check_pbc(pbc, structure):   
    if pbc is None:
        return all(structure.get_pbc())
    else:
        assert(isinstance(pbc, bool))
        return pbc


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Set of functions to work with Boys and give us amazing orbitals.

Created on Thu Jul 13 15:55:09 2023

@author: roncoroni
"""

from pyscf import lo
#%%

def do_localization_pyscf(dft_calc, deg_orbitals, loc_type):
    if "boys" in loc_type.lower():
        mo_loc = do_boys_pyscf(dft_calc, deg_orbitals)
    elif "ibo" in loc_type.lower():
        mo_loc = do_ibo_pyscf(dft_calc, deg_orbitals)
        
    return mo_loc
        
def do_boys_pyscf(dft_calc, deg_orbitals):

    mo_boys = dft_calc.mo_coeff.copy()
    for ii in [0,1]:

        # make loc object
        loc = lo.Boys(dft_calc.mol)

        # run boys on MO from calculation
        loc.run(mo_coeff=dft_calc.mo_coeff[ii][:, deg_orbitals[ii]])

        # reupdate old orbitals with boys
        mo_boys[ii][:, deg_orbitals[ii]] = loc.mo_coeff

    return mo_boys

def do_ibo_pyscf(dft_calc, deg_orbitals):

    mo_ibo = dft_calc.mo_coeff.copy()
    for ii in [0, 1]:
        pm = lo.ibo.ibo(dft_calc.mol, dft_calc.mo_coeff[ii][:, deg_orbitals[ii]])

        mo_ibo[ii][:, deg_orbitals[ii]] = pm
    
    return mo_ibo


"""
*****************************************************
| PyMBXAS: Python-based MBXAS implementation        |
*****************************************************

PyMBXAS is a package for setting up, manipulating, running and visualizing MBXAS calculations using Python. It has an object-oriented approach to simplify the task of spectra analysis and post-processing.
PyMBXAS leverages the PySCF  electronic structure code and the Atomic Simulation Environment (ASE).

The current implementation is being developed on:
    https://gitlab.com/roncofaber/pymbxas

The initial MBXAS work is based on:
    https://github.com/subhayanroychoudhury/CleaRIXS
    
If you want more information, please look at the following publications:
    
    1. https://journals.aps.org/prb/abstract/10.1103/PhysRevB.106.115115
    2. https://journals.aps.org/prb/abstract/10.1103/PhysRevB.107.035146

Example script:
    
    import ase 
    import ase.build
    from pymbxas.calculators.pyscf import PySCF_mbxas

    # read structure in ASE format
    structure = ase.build.molecule("H2O")

    # set up calculation parameters
    channel   = 1 
    charge    = 0 # charge of system
    spin      = 0 # spin of system
    to_excite = "O" # index(es)/symbols of atom(s) to excite
    basis     = "def2-svpd"
    xc        = "lda"#"b3lyp"
    pbc       = False

    # set up object
    obj = PySCF_mbxas(
        structure    = structure,
        charge       = charge,
        spin         = spin,
        xc           = xc, 
        basis        = basis,
        do_xch       = True, # do XCH for energy alignment
        target_dir   = None, # specify a target directory where to run calc
        verbose      = 3,     # level of verbose of pySCF output (4-5 is good)
        print_fchk   = False,  # print fchk files (requires MOKIT)
        print_output = False, # print output to console
        save         = True,  # save object as pkl file
        save_name    = "pyscf_obj.pkl", # name of saved file
        save_path    = None, # path of saved object
        loc_type     = "ibo",
        gpu          = True # if you want to use the GPU code
        )

    # run calculation (GS + FCH + XCH)
    obj.kernel(to_excite)

    # plot result
    E, I = obj.get_mbxas_spectra(to_excite)#, erange=[520, 540], sigma=0.006)

    plt.figure()
    plt.plot(E,I)
    plt.show()

"""

# import os
import sys

if sys.version_info[0] == 2:
    raise ImportError('Please run with Python3. This is Python2.')

# package info
__version__ = '0.4.1'
__date__ = "15 Oct. 2024"
__author__ = "Fabrice Roncoroni"
__all__ = ["spectra", "spectras"]

from pymbxas.spectra  import Spectra
from pymbxas.spectras import Spectras



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  2 15:32:48 2024

@author: roncofaber
"""

# numpy is my rock and I am ready to roll
import numpy as np

# sea urchin, maybe not needed in the future #TODO
import sea_urchin.clustering.metrics as met
import sea_urchin.clustering.clusterize as clf

# pymbxas stuff
from pymbxas.explorer.node import spectralNode
from pymbxas.mbxas.broaden import get_mbxas_spectra

# learning stuff
from sklearn.gaussian_process import GaussianProcessRegressor
import sklearn.gaussian_process.kernels as sker

# instead of random floats, let's make some order
from ase import units
Ha = units.Ha

#%%

# wrapper class to predict a spectra
class MBXASplorer():
    
    def __init__(self, spectras, fit_labels=None, xscaler="standard",
                 yscaler="standard", metric=None, isotropic=True, verbose=0,
                 ykernel=None):
        
        # set up internal variables
        self._verbose = verbose
        self._is_isotropic = isotropic
        self._ykernel = ykernel
        
        self._nodes  = []
        self._labels = []
        
        # define labels to fit
        if fit_labels is None:
            fit_labels = np.unique([sp._el_labels for sp in spectras])
        else:
            fit_labels = [ii for ii in np.unique(fit_labels)
                          if ii in np.unique([sp._el_labels for sp in spectras])]
            
        # define metric (must work on ASE atoms object)
        if metric is None:
            metric = met.get_distances
        self._metric = metric
        
        # generate scaler for feature vector
        self.xscale = clf.generate_scaler(xscaler)
        
        # calculate feature vector for all spectra
        structures = [sp.structure for sp in spectras]
        X  = self.metric(structures)
        Xs = self.xscale.fit_transform(X)
        
        # read GS data
        ys_G = self._read_scale_GS_energy(spectras, yscaler)
        
        # first, predict GS energy
        self.kr_g = self._fit_gs_energy(Xs, ys_G)

        # fit each peak
        for peak_label in fit_labels:
            self._make_node(spectras, peak_label, Xs, yscaler, isotropic, ykernel)
            
        return
    
    def _make_node(self, spectras, peak_label, Xs, yscaler, isotropic, ykernel):
        
        if peak_label == -1 or peak_label in self._labels: # ignore noise
            return
        
        # fit a single excitation
        sn = spectralNode(spectras, peak_label, Xs, yscaler=yscaler,
                          isotropic=isotropic, ykernel=ykernel)
          
        # add it to the list #TODO: add later as well
        self._nodes.append(sn)
        self._labels.append(peak_label)
        
        # if self._verbose:
            # print(sn.kr_a.kernel_)
                
        return
    
    def predict(self, structures, node=None, return_std=False):
        
        if not isinstance(structures, list): # it's a single structure
            Xtest = self.metric([structures])
        else:
            Xtest = self.metric(structures)
        
        # scale new input
        Xtest_scaled = self.xscale.transform(Xtest)
        
        y_G = self._predict_gs_energy(Xtest_scaled)
        
        if node is not None:
            
            E_pre, E_std, A_pre, A_std = self[node].predict(Xtest_scaled)
        
        # iterate over every node
        else:
            
            E_pre, E_std, A_pre, A_std = [], [], [], []
            for node in self:
                e_pre, e_std, a_pre, a_std = node.predict(Xtest_scaled)
                
                E_pre.append(e_pre.reshape(-1, node.n_targets))
                E_std.append(e_std.reshape(-1, node.n_targets))
                A_pre.append(a_pre.reshape(-1, node.n_targets))
                A_std.append(a_std.reshape(-1, node.n_targets))
                
            E_pre = np.hstack(E_pre)
            E_std = np.hstack(E_std)
            A_pre = np.hstack(A_pre)
            A_std = np.hstack(A_std)
                
        y_G = np.squeeze(y_G)
        
        if return_std:
            return E_pre, E_std, A_pre, A_std, y_G
        else:
            return E_pre, A_pre, y_G
            
    
    # generate a MBXAS spectra predicted from a structure
    def get_mbxas_spectra(self, structure, axis=None, sigma=0.02, npoints=3001,
                          tol=0.01, erange=None):
        
        # predict E and A
        energies, amplitude, __ = self.predict(structure)
        
        # get broadened spectra
        erange, spectras = get_mbxas_spectra(energies, amplitude,
                                             sigma=sigma, npoints=npoints,
                                             tol=tol, erange=erange,
                                             isotropic=self._is_isotropic)

        if isinstance(structure, list):
            norm = len(structure)
        else:
            norm = 1
        
        #return
        if axis is None:
            spectras = np.mean(spectras, axis=0)/norm
        else:
            spectras = spectras[axis]/norm
        
        return erange, spectras
    
    def _read_scale_GS_energy(self, spectras, yscaler):
        
        y_G = np.array([sp.gs_energy for sp in spectras])
  
        self.yscale_G = clf.generate_scaler("none")
        
        ys_G = self.yscale_G.fit_transform(y_G.reshape(-1, 1))
        
        return ys_G
    
    @staticmethod
    def _fit_gs_energy(Xs, ys_G):
        
        kernel = 3.0 * sker.RBF(
            length_scale        = 15.0,
            length_scale_bounds = (1e-3, 1e3)
            )
        
        kr_g = GaussianProcessRegressor(
            kernel,
            n_restarts_optimizer = 30
            ).fit(Xs, ys_G)

        return kr_g
    
    def _predict_gs_energy(self, Xtest):
        gs_energy = self.kr_g.predict(Xtest).reshape(-1, 1)
        return Ha*self.yscale_G.inverse_transform(gs_energy)
    
    # make class iterable
    def __getitem__(self, index):
        return self._nodes[index]
    
    def __iter__(self):
        return iter(self._nodes)
    
    def __len__(self):
        return len(self.nodes)
    
    @property
    def nodes(self):
        return self._nodes

    def metric(self, X):
        return self._metric(X)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug  3 15:29:31 2023

@author: roncoroni
"""

import numpy as np

import ase

from pyscf import gto
from pyscf.pbc import gto as pgto

from pymbxas.io.logger import Logger
from pymbxas.utils.check_keywords import check_pbc
from pymbxas.utils.auxiliary import get_available_memory

try:
    import sea_urchin.alignement.align as ali
    has_SU = True
except:
    has_SU = False
#%%

# convert an ase Atoms object to a mole or cell object for pyscf
def ase_to_mole(structure, charge=0, spin=0, basis='def2-svpd', pbc=None,
                verbose=4, print_output=True, symmetry=False,
                is_gpu=False, **kwargs):

    # generate atom list to feed to object
    atom_list = []
    for ii in range(len(structure)):
        atom_list.append([
            structure.get_chemical_symbols()[ii],
            tuple(structure.get_positions()[ii])
            ])
    
    if pbc is None:
        pbc = check_pbc(pbc, structure)
    
    # periodic system
    if pbc:
        
        mol = pgto.Cell(
            atom  = atom_list,
            basis = basis,
            charge = charge,
            spin = spin,
            verbose = verbose,
            stdout = Logger(print_output),
            a = structure.get_cell().array,
            ke_cutoff = 100.0,
            symmetry = symmetry,
            **kwargs
            )
    
    # non periodic system
    else:

        mol = gto.Mole(
            atom  = atom_list,
            basis = basis,
            charge = charge,
            spin = spin,
            verbose = verbose,
            stdout = Logger(print_output),
            max_memory = get_available_memory(is_gpu),
            unit = 'Angstrom',
            symmetry = symmetry,
            **kwargs
            )

    mol.build()
    
    # overwrite intergrals if PBC with proper ones #TODO: check if needed and correct
    if pbc:
        mol.intor = mol.pbc_intor
    
    return mol

# convert a mol object to ase Atoms
def mole_to_ase(mol, units="Angstrom", **kwargs):
    
    structure = ase.Atoms(
        mol.elements,
        mol.atom_coords(unit=units),
        **kwargs
        )
    
    return structure

# rotate, translate, permute, inverse a structure
def rotate_structure(structure, rot, tr, perm, inv, rtype):
    assert has_SU, "Please install Sea Urchin to use this"
    return ali.align_structure(structure, rot, tr, perm, inv, rtype)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  2 15:33:50 2024

@author: roncofaber
"""

import numpy as np

import sea_urchin.clustering.clusterize as clf

# from sklearn.gaussian_process import GaussianProcessRegressor
# import sklearn.gaussian_process.kernels as sker

import pymbxas.utils.auxiliary as aux

import copy

import gpflow

# from ase import units
# Ha = units.Ha

#%%

# class of a single electronic cluster
class spectralNode():
    
    def __init__(self, spectras, label, Xdata, yscaler="standard",
                 isotropic=False, ykernel=None):
        
        # read data to use for fitting
        Xs, y_E, y_A, n_targets = self._read_data(spectras, label, Xdata, isotropic)
        
        self.y_A       = y_A
        self.n_targets = n_targets
        
        # scale data accordinglspectray
        ys_E, ys_A = self._scale_data(y_E, y_A, yscaler)
        
        # do fitting for energies and amplitudes
        self.kr_e = self._fit_energy(Xs, ys_E, n_targets, ykernel)
        self.kr_a = self._fit_amplitudes(Xs, ys_A, n_targets, isotropic, ykernel)
        
        # assign label variable
        self.label = label
            
        return
    
    # read data from spectra and return them
    @staticmethod
    def _read_data(spectras, label, Xdata, isotropic):
        
        # obtain number of targets
        n_targets = int(aux.standardCount([sp._el_labels for sp in spectras], label))
        
        if n_targets == 0:
            print("WARNING: 0")
            n_targets = 1
        
        # read spectral data
        y_E  = []
        y_A  = []
        Xout = []
        for cc, spectra in enumerate(spectras):
            
            idxs = np.where(spectra._el_labels == label)[0]
            
            # ignore if wrong number of targets
            if len(idxs) != n_targets:
                continue
    
            # append energies
            y_E.append(spectra.energies[idxs])
            
            # append values to fit amplitude
            amp = spectra.amplitude[:,idxs]
            y_A.append(amp**2) # append square value of the amplitude
            
            # store training coordinates
            Xout.append(Xdata[cc])
        
        # define values for fitting (convert to eV)
        y_E  = np.array(y_E).reshape(-1, n_targets)
        Xout = np.array(Xout) 
        y_A  = np.array(y_A)
        
        if isotropic:
            # y_A = np.mean(y_A, axis=1).reshape(-1, 1, n_targets)
            y_A = np.mean(y_A, axis=1).reshape(-1, n_targets)
        
        return Xout, y_E, y_A, n_targets
    
    # take read data and return scaled data while generating the scalers
    def _scale_data(self, y_E, y_A, yscaler):
        
        # generate data scalers
        self.yscale_E = clf.generate_scaler(yscaler)
        self.yscale_A = clf.generate_scaler(yscaler)
      
        # scale 'em
        ys_E = self.yscale_E.fit_transform(y_E)
        ys_A = self.yscale_A.fit_transform(y_A)
        
        self._std_E = np.sqrt(self.yscale_E.var_)
        self._std_A = np.sqrt(self.yscale_A.var_)
        
        return ys_E, ys_A
    
    @staticmethod
    def _fit_energy(Xs, ys_E, n_targets, ykernel):
        
        # if ykernel is None:
        #     kernel = 5.0 * sker.RBF(
        #         length_scale        = 15.0,
        #         length_scale_bounds = (1e-3, 1e3)
        #         )
        # else:
        #     kernel = copy.deepcopy(ykernel)
        
        # k_e = GaussianProcessRegressor(
        #     kernel,
        #     n_restarts_optimizer = 25,
        #     n_targets            = n_targets
        #     ).fit(Xs, ys_E)
        
        model_E = gpflow.models.GPR(
            (Xs, ys_E),
            kernel=gpflow.kernels.SquaredExponential(),
        )
        
        opt_E = gpflow.optimizers.Scipy()
        opt_E.minimize(model_E.training_loss, model_E.trainable_variables)
        
        return model_E
    
    @staticmethod
    def _fit_amplitudes(Xs, ys_A, n_targets, isotropic, ykernel):
        
        assert isotropic, "Only isotropic implemented"
        
        
        # if ykernel is None:
        #     kernel = sker.RBF() + sker.Matern()
        # else:
        #     kernel = copy.deepcopy(ykernel)

        # k_a = GaussianProcessRegressor(kernel,
        #                                n_restarts_optimizer = 10,
        #                                n_targets            = n_targets
        #                                ).fit(Xs, ys_A)
        
        model_A = gpflow.models.GPR(
            (Xs, ys_A),
            kernel=gpflow.kernels.SquaredExponential(),
        )
        
        opt_A = gpflow.optimizers.Scipy()
        opt_A.minimize(model_A.training_loss, model_A.trainable_variables)
        
        return model_A
    
    def _predict_energy(self, Xtest):
        
        e_pre, e_std = self.kr_e.predict_f(Xtest)
        
        e_pre = e_pre.numpy().reshape(-1, self.n_targets)
        e_std = self._std_E*e_std.numpy()
        
        return np.squeeze(self.yscale_E.inverse_transform(e_pre)), np.squeeze(e_std)
    
    def _predict_amplitude(self, Xtest):
        
        # xas_amplitudes = []
        # for kr_a in self.kr_a:
        #     xas_amplitude = kr_a.predict(Xtest)
        #     xas_amplitudes.append(xas_amplitude)
        
        a_pre, a_std = self.kr_a.predict_f(Xtest)
        
        # xas_amplitudes = np.concatenate(xas_amplitudes).reshape(-1,self.n_targets)
        a_pre = a_pre.numpy().reshape(-1, self.n_targets)
        a_std = self._std_A*a_std.numpy()
        
        return np.squeeze(self.yscale_A.inverse_transform(a_pre)), np.squeeze(a_std)
    
    
    def predict(self, Xtest_scaled):
        
        e_pre, e_std = self._predict_energy(Xtest_scaled)
        a_pre, a_std = self._predict_amplitude(Xtest_scaled)
        
        return e_pre, e_std, a_pre, a_std



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Set of functions to work with Boys and give us amazing orbitals.

Created on Thu Jul 13 15:55:09 2023

@author: roncoroni
"""

import numpy as np
# import cupy as cp

from .basis import get_basis_set_info
#%%

# function to generate a list of 1s orbitals for a specific channel
def find_1s_in_channel(boys_coeff, atom_coeffs, atom_labels, symbols):
    symbol_list = np.concatenate([[cc]*len(atom_coeffs[cc]) for cc in range(len(symbols))])
    labels_list = np.concatenate(atom_labels)
    dominant_atoms = np.argmax(np.abs(boys_coeff), axis=1)

    orb_types   = labels_list[dominant_atoms]
    orb_symbols = symbol_list[dominant_atoms]

    found_elements = []
    s1_list = [None]*len(symbols)
    for idx, (orb_typ, orb_id) in enumerate(zip(orb_types, orb_symbols)):

        if orb_typ == "s" and orb_id not in found_elements:
            found_elements.append(orb_id)
            s1_list[orb_id] = idx
    return s1_list

# find the 1s orbitals of the system
def find_1s_orbitals(gs_data, use_localized=False):

    # get basis set information
    atom_coeffs, atom_labels, symbols, nbasis, indexing = get_basis_set_info(gs_data['basis'])

    s_orbitals   = {}
    # iterate over channels and find 1s orbitals
    for channel in ["alpha", "beta"]:

        if use_localized:
            orb_coeff = np.array(gs_data["localized_coefficients"][channel])
        else:
            orb_coeff = np.array(gs_data["coefficients"][channel])

        # calculate which ones are the 1s orbitals
        s1_list = find_1s_in_channel(orb_coeff, atom_coeffs, atom_labels, symbols)
        s_orbitals[channel] = s1_list

    return s_orbitals

# calculate the boys overlap
def calculate_boys_overlap(boys_coeffs, fch_coeffs, basis_overlap):

    boys_overlap = {}
    # iterate over channels and calculate overlap
    for channel in ["alpha", "beta"]:

        boys_coeff = np.array(boys_coeffs[channel])
        exci_coeff = np.array(fch_coeffs[channel])

        # calculate boys overlap factor
        OVLP = np.linalg.multi_dot([boys_coeff, basis_overlap, exci_coeff.T])
        boys_overlap[channel] = OVLP

    return boys_overlap


def find_1s_orbitals_pyscf(molecule, coefficients, energies, occupations,
                           ato_idxs, check_deg=True):

    # occupied orbitals only
    occ_idxs = np.where(occupations == 1)[0]
    
    # orbital labels
    ao_labels = np.array(molecule.ao_labels(fmt=False))

    orb_list = []
    for to_excite in ato_idxs:
        
        symbol = molecule.atom_symbol(to_excite)
    
        for cc, orb in enumerate(coefficients.T[occ_idxs]):
    
            # square coeff
            orb2 = orb**2
    
            # find indexes where orbital has weight
            rel_idxs = np.where(orb2 > 0.3*np.max(orb2))[0]
            
            # if isinstance(rel_idxs, cp.ndarray):
            #     rel_idxs = rel_idxs.get()
    
            rel_labels = ao_labels[rel_idxs]
            # there is weight on our orbital of interest
            if any([all(lab == (to_excite, symbol, "1s", "")) for lab in rel_labels]):
    
                # if not in list add it and all the orbitals with similar energy
                if cc in orb_list:
                    continue
    
                if check_deg:
                    degenerate_orbs = np.where(np.abs(energies - energies[cc]) < 1e-1)[0].tolist()
                    orb_list.extend(degenerate_orbs)
                else:
                    orb_list.append(cc)

    return orb_list


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 29 13:17:41 2024

@author: roncoroni
"""

import time
import copy

# self module utilities
from pymbxas.io.data import pyscf_data
from pymbxas.build.structure import ase_to_mole
from pymbxas.build.input_pyscf import make_pyscf_calculator
from pymbxas.utils.orbitals import find_1s_orbitals_pyscf
from pymbxas.mbxas.mbxas import run_MBXAS_pyscf

# pyscf stuff
from pyscf.scf.addons import mom_occ

# MOKIT stuff
from pymbxas.io.write import write_data_to_fchk
    
#%%
    
class Excitation(object):
    
    def __init__(self, structure, gs_data, ato_idx, parameters, channel,
                 df_obj, oset, logger):
        
        # set up excitation info
        self.ato_idx   = ato_idx
        self.symbol    = structure.get_chemical_symbols()[ato_idx]
        self.channel   = channel
        
        # store output
        self.output = {}
        self.data   = {}
        
        # find index of orbital to excite
        orb_idx = find_1s_orbitals_pyscf(gs_data.mol, gs_data.mo_coeff[channel],
                                         gs_data.mo_energy[channel],
                                         gs_data.mo_occ[channel],
                                         [ato_idx],
                                         check_deg=False)
        
        # check that the orbitals are not still delocalized
        if not len(orb_idx) == 1:
            logger.error("It seems that the atomic orbitals are still delocalized.")
            logger.error("Excitation of {:<2} atom #{:>2} aborted.".format(
                self.symbol, self.ato_idx))
            return
        
        # assign index of the orbital to excite
        self.orb_idx = orb_idx[0]
        
        # make it work even if it uses GPU
        if oset["is_gpu"]:
            gs_data = gs_data.to_gpu()
            
        # run excitation
        self._excite(structure, gs_data, parameters, df_obj, logger, oset)
        
        return
    
    
    # use it to run excitation of the selected atom
    def _excite(self, structure, gs_data, parameters, df_obj, logger, oset):
        
        logger.info("-----> Exciting {:<2} atom #{:>2} <-----|".format(
            self.symbol, self.ato_idx))
        
        # run FCH
        self._run_fch(structure, gs_data, parameters, df_obj, logger, oset)
        
        # run XCH
        self._run_xch(structure, gs_data, parameters, df_obj, logger, oset)
        
        # run MBXAS
        self._run_mbxas(gs_data, logger)
        
        logger.info("----- Excitation successful! -----|\n")
        return
    
    # run the FCH calculation
    def _run_fch(self, structure, gs_data, parameters, df_obj, logger, oset):
        
        logger.info(">>> Started FCH calculation.")
        
        # retrieve parameters
        pbc       = parameters["pbc"]
        charge    = parameters["charge"] + 1 if not pbc else 0 #TODO: check if works properly for PBC
        spin      = parameters["spin"] + self.channel*2 - 1 if not pbc else 0 #TODO: check if works properly for PBC
        basis     = parameters["basis"]
        xc        = parameters["xc"]
        solvent   = parameters["solvent"]
        calc_type = parameters["calc_type"]

        start_time  = time.time()

        # Read MO coefficients and occupation number from GS
        scf_guess  = copy.deepcopy(gs_data.mo_coeff)
        occupation = copy.deepcopy(gs_data.mo_occ)

        # Assign initial occupation pattern --> kick orbital N
        occupation[self.channel][self.orb_idx] = 0
        
        # assign magmom
        magmom = len(structure)*[0]
        magmom[self.ato_idx] = 1-2*self.channel

        # change charge
        fch_mol = ase_to_mole(structure, charge=charge, spin=spin, basis=basis,
                              pbc=pbc, verbose=oset["verbose"],
                              print_output=oset["print_output"],
                              magmom=magmom, is_gpu=oset["is_gpu"])

        # Defnine new SCF calculator
        fch_calc = make_pyscf_calculator(fch_mol, xc=xc, calc_type=calc_type,
                                         pbc=pbc, solvent=solvent,
                                         dens_fit=df_obj, calc_name="fch",
                                         save=oset["save_chk"],
                                         gpu=oset["is_gpu"],)

        # Construct new density matrix with new occupation pattern
        dm_u = fch_calc.make_rdm1(scf_guess, occupation)

        # Apply mom occupation principle
        fch_calc = mom_occ(fch_calc, scf_guess, occupation)

        # Start new SCF with new density matrix
        fch_calc.scf(dm_u)

        # store input/output
        self.output["fch"] = fch_calc.stdout.log.getvalue()
        self.data["fch"]   = pyscf_data(fch_calc)

        # if pobj._print_fchk:
        #     write_data_to_fchk(pobj.mol, fch_calc,
        #                        oname = pobj._tdir + "/output_fch_{}.fchk".format(
        #                            self.ato_idx),
        #                        )
            
        logger.info(">>> FCH finished in {:.1f} s.".format(time.time() - start_time))
        return

    # run the XCH calculation
    def _run_xch(self, structure, gs_data, parameters, df_obj, logger, oset):
        
        logger.info(">>> Started XCH calculation.")
        
        # retrieve parameters
        pbc       = parameters["pbc"]
        charge    = parameters["charge"] 
        spin      = parameters["spin"]
        basis     = parameters["basis"]
        xc        = parameters["xc"]
        solvent   = parameters["solvent"]
        calc_type = parameters["calc_type"]
        
        if oset["is_gpu"]:
            data = self.data["fch"].to_gpu()
        else:
            data = self.data["fch"]
        
        start_time = time.time()

        # Read MO coefficients and occupation number from GS
        scf_guess  = copy.deepcopy(data.mo_coeff)
        occupation = copy.deepcopy(data.mo_occ)

        # Assign initial occupation pattern --> kick orbital N
        # occupation[channel][self.orb_idx] = 0
        occupation[self.channel][gs_data.nelec[self.channel]] = 1

        # make XCH molecule
        xch_mol = ase_to_mole(structure, charge=charge, spin=spin, basis=basis,
                              pbc=pbc, verbose=oset["verbose"],
                              print_output=oset["print_output"],
                              is_gpu=oset["is_gpu"])

        # define new SCF calculator
        xch_calc = make_pyscf_calculator(xch_mol, xc=xc, calc_type=calc_type,
                                         pbc=pbc, solvent=solvent,
                                         dens_fit=df_obj, calc_name="xch",
                                         save=oset["save_chk"],
                                         gpu=oset["is_gpu"])

        # Construct new density matrix with new occupation pattern
        dm_u = xch_calc.make_rdm1(scf_guess, occupation)

        # Apply mom occupation principle
        xch_calc = mom_occ(xch_calc, scf_guess, occupation)

        # Start new SCF with new density matrix
        xch_calc.scf(dm_u)
        
        # reconvert to cpu cause otherwise stuff is iffy
        # if self.oset["is_gpu"]:
            # xch_calc = xch_calc.to_cpu()

        # store input/output
        self.output["xch"] = xch_calc.stdout.log.getvalue()
        self.data["xch"]   = pyscf_data(xch_calc)

        # if pobj._print_fchk:
        #     write_data_to_fchk(pobj.mol, xch_calc,
        #                        oname = pobj._tdir + "/output_xch_{}.fchk".format(
        #                            self.ato_idx),
        #                        )

        logger.info(">>> XCH finished in {:.1f} s.".format(time.time() - start_time))
        
        return
    
    # run MBXAS from a set of pySCF calculations
    def _run_mbxas(self, gs_data, logger):
        
        start_time = time.time()

        energies, absorption, mb_ovlp, dip_KS, b_ovlp = run_MBXAS_pyscf(
            gs_data.mol, gs_data.to_cpu(), self.data["fch"].to_cpu(),
            self.orb_idx, channel=self.channel, xch_calc=self.data["xch"].to_cpu())

        self.mbxas = {
            "energies"   : energies,
            "absorption" : absorption,
            "mb_overlap" : mb_ovlp,
            "dipole_KS"  : dip_KS,
            "basis_ovlp" : b_ovlp
            }
        
        logger.info(">>> MBXAS finished in {:.1f} s [\u2713].".format(time.time() - start_time))
        return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul  7 13:09:15 2023

@author: roncofaber
"""

import os
import glob
import shutil

#%%

def copy_output_files(source_dir, target_dir):
    try:
        if not os.path.exists(target_dir):
            os.mkdir(target_dir)
            
        file_list = glob.glob(source_dir + "/*.txt")
        for file in file_list:
            shutil.copy(file, target_dir)
    except OSError:
        print("Copy failed!")
        pass
    
    return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Set of functions to set up AIMD and relaxations

Created on Wed Mar 13 10:35:07 2024

@author: roncofaber
"""

import ase
import ase.visualize

from pyscf.md.distributions import MaxwellBoltzmannVelocity
import pyscf.geomopt

from pymbxas.build.structure import ase_to_mole
from pymbxas.build.input_pyscf import make_pyscf_calculator
from pymbxas.md.callback import OptimizerTraj, AIMDTrajWriter
from pymbxas.md.integrator import NVTBerendson

#%%

class Geometry_optimizer():
    """Class to handle geometry optimization using PySCF.

    Args:
        pyscf_calc (pyscf object): A PySCF calculation object.
        opt_type (str, optional): Type of optimization algorithm ('geometric' or 'berny'). Defaults to 'geometric'.
        write (bool, optional): Whether to write trajectory to a file. Defaults to True.
        oname (str, optional): Output filename. Defaults to "optimization.xyz".
        append (bool, optional): Whether to append to an existing file. Defaults to False.
        density_fit (bool, optional): Whether to use density fitting for the calculation. Defaults to False.
    """
    
    def __init__(self, pyscf_calc, opt_type="geometric",
                 write=True, oname="optimization.xyz", append=False,
                 density_fit=False):
        
        self.opt_type = opt_type
        
        # make dens. fitting
        if density_fit:
            pyscf_calc.density_fit()
            
        # assign calculator
        self.calc = pyscf_calc
        
        # generate optimizer object
        self.opt = OptimizerTraj(write=write, oname=oname, append=append)

        return
    
    def optimize(self, maxsteps=100):
        
        optimize = self._get_opt_type()
        
        # optimize mol geometry
        self.mol_eq = optimize(self.calc, maxsteps=maxsteps, callback=self.opt)
        
        return
    
    def _get_opt_type(self):
        
        # check solver type
        if self.opt_type == "geometric":
            from pyscf.geomopt.geometric_solver import optimize
        elif self.opt_type == "berny":
            from pyscf.geomopt.berny_solver import optimize
        else:
            raise "WRONG 'opt_type' selected. 'geometric' and 'berny' accepted."
        
        return optimize
    
    def view(self):
        self.opt.view()
        return
    
    # make class iterable
    def __getitem__(self, index):
        return self.opt[index]
    
    def __iter__(self):
        return iter(self.opt._traj)
    
    # get elements
    @property
    def mol(self):
        return self.opt.mol
    @property
    def traj(self):
        return self.opt.traj
    @property
    def atoms(self):
        return self.opt.atoms
        
    

class AIMD_solver():
    """
    """
    def __init__(self, pyscf_calc, T, taut=100, dt=10, nsteps=1, ofreq=1,
                 oname="aimd_traj.xyz", veloc=None, output="output.log",
                 fixcom=True, fixrot=True):
        
        # run GS calc
        if not pyscf_calc.converged:
            pyscf_calc.kernel()
        
        # initialize velocity
        if veloc == "rand":
            veloc = MaxwellBoltzmannVelocity(pyscf_calc.mol, T=T)

        # make a logger
        self.logger = AIMDTrajWriter(
            oname        = oname, # name output traj
            nstep        = ofreq,    # frequency at which traj is written
            data_output  = output, # also write output
            append       = False # append to existing traj
            )
        
        # make scanner
        scanner = pyscf_calc.as_scanner()

        self.integrator = NVTBerendson(scanner,
                                  dt    = dt,
                                  steps = nsteps,
                                  T     = T,
                                  taut  = taut,
                                  veloc =  veloc,
                                  callback = self.logger,
                                  fixrot = fixrot,
                                  fixcom = fixcom
                                  )
        
        return
    
    # run AIMD
    def run(self, veloc=None, steps=None, dump_flags=True, verbose=None):
        self.integrator.kernel(veloc=veloc, steps=steps,
                               dump_flags=dump_flags, verbose=verbose)
        return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 29 14:06:44 2024

@author: roncoroni
"""

    def _calculate_livvo(self, locmethod='IBO', iaos=None, s=None,
              exponent=4, grad_tol=1e-8, max_iter=200, verbose=None): #TODO only for channel 1 (but should be same)
        
        mo_coeff = self.data.mo_coeff[1].copy()
        mo_occ   = self.data.mo_occ[1].copy()
        
        occ_idxs = np.where(mo_occ > 0)[0]
        uno_idxs = np.where(mo_occ == 0)[0]
        
        mo_vvo = lo.vvo.livvo(self.mol,
                              mo_coeff[:, occ_idxs], mo_coeff[:, uno_idxs],
                              locmethod=locmethod,
                              iaos=iaos, s=s,
                              exponent=exponent, grad_tol=grad_tol,
                              max_iter=max_iter, verbose=verbose)
        
        return mo_vvo
    
    
    def get_excited_orbital_character(self, ato_idx, livvo=None):
        
        if livvo is None:
            livvo = self.data.mo_livvo
            
        basis_ovlp = self.mol.intor("int1e_ovlp")
        
        ato_idxs = atoms_to_indexes(self.structure, ato_idx)
        
        projections = []
        for excitation in self.excitations:
            if excitation.ato_idx not in ato_idxs:
                continue
            
            data = excitation.data["fch"]
            
            # get unoccupied orbitals
            uno_idxs = np.where(data.mo_occ[excitation.channel] == 0)[0][1:]
            CMO      = data.mo_coeff[excitation.channel][:, uno_idxs]
            
            # calculate projection of CMOs on LIVVOs basis
            projection = (CMO.conj().T@basis_ovlp@livvo)**2
            
            projections.append(projection)
        
        if len(ato_idxs) == 1:
            return projections[0]
        else:
            return projections


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  5 12:08:17 2023

@author: roncofaber
"""

import numpy as np
import copy
import dill

from pymbxas import Spectra

# can we use the sea urchin here?
try:
    # import sea_urchin.alignement.align as ali
    import sea_urchin.clustering.metrics as met
    import sea_urchin.alignement.align as ali
    SeaUrchin_exists = True
except:
    SeaUrchin_exists = False
    
#%%

"""
Class for a collection of spectras - WIP

Input: list of pyscf objects

"""

class Spectras():
    
    def __init__(self, spectra_list,
                 labels     = None,
                 post_align = False,
                 alignment  = None
                 ):
        
        if isinstance(spectra_list, list):
            self.__initialize_collection(spectra_list, labels, post_align,
                                     alignment)
        else:
            self.__restart(spectra_list)
        
        return
    
    # start from list of pyscf objects
    def __initialize_collection(self, spectra_list, labels, post_align,
                                alignment):
        
        # copy input spectras
        self.spectras = copy.deepcopy(spectra_list)
        
        # assign labels, if existend
        self.assign_atomic_labels(labels)
        
        # align
        if post_align:
            assert alignment is not None, "Provide an alignment"
            self.align_labels_to_mean_structures(alignment)
            
        return
    
    # restart from pkl file or another spectras object
    def __restart(self, spectra_obj):
        
        #load a pkl or dict
        if isinstance(spectra_obj, dict):
            self.__dict__ = spectra_obj.copy()
        elif isinstance(spectra_obj, str):
            self.__dict__ = self.__pkl_to_dict(spectra_obj)
            
        spectras = [Spectra(ii) for ii in self.spectras]
        
        self.spectras = spectras
            
        return
    
    def __pkl_to_dict(self, filename):
        with open(filename, 'rb') as fin:
            data = dill.load(fin)
        return data
    
    def assign_atomic_labels(self, labels):
        
        # check if labels are provided
        if labels is None:
            self.labels = len(self.spectras)*[-1]
        else:
            self.labels = labels.copy()
                
        # check that dims match
        assert len(self.labels) == len(self.spectras), "Wrong labels"
        
        # assign labels
        for cc, spectra in enumerate(self.spectras):
            spectra._label = self.labels[cc]

        # reset aligned keyword        
        self._aligned = False
        
        return
    
    def assign_electronic_labels(self, labels, label=None):
        
        if label is None:
            spectras = self.spectras
        else:
            spectras = self.__get_atomic_label(label)
        
        # check is all good
        assert len(labels) == len(spectras)
        
        imax = labels.shape[1]
        
        # assign electronic labels
        for cc, sp in enumerate(spectras):
            tlab = -np.ones(len(sp.energies), dtype=int)
            tlab[:imax] = labels[cc]
            
            sp._el_labels = tlab
        
        return
    
    # return sliced object with specific label
    def get_spectra_with_label(self, label):
        
        sp_list = self.__get_atomic_label(label)
        
        return Spectras(sp_list, labels=len(sp_list)*[label])
    
    # get all spectras with a specific label
    def get_mbxas_spectras(self, axis=None, sigma=0.02, npoints=1001, tol=0.01,
                          erange=None, label=None, el_label=None):
        if label is None:
            spectras = self.spectras
        else:
            spectras = self.__get_atomic_label(label)
        
        I_list = []
        for spectra in spectras:
            E, I = spectra.get_mbxas_spectra(axis=axis, sigma=sigma,
                                             npoints=npoints, tol=tol,
                                             erange=erange, el_label=el_label)
            I_list.append(I)
        
        return E, np.array(I_list)
   
    # get the average spectra
    def get_mbxas_spectra(self, axis=None, sigma=0.02, npoints=1001, tol=0.01,
                          erange=None, label=None, el_label=None):
        
        E, I_list = self.get_mbxas_spectras(axis=axis, sigma=sigma,
                                            npoints=npoints, tol=tol,
                                            erange=erange, label=label,
                                            el_label=el_label)
        
        return E, np.mean(I_list, axis=0)
    
    def align_labels_to_mean_structures(self, alignment):
        

        for lab in set(self.labels):
            
            if lab == -1: #ignore noise
                continue
            
            self._align_label_to_mean_structure(lab, alignment)
            
        self._aligned = True
        
        return
    
    def _align_label_to_mean_structure(self, label, alignment):
        
        # get spectras and structures
        spectras   = self.__get_atomic_label(label)
        structures = [sp.structure for sp in spectras]
        
        # calculate mean structure
        __, mstrus = ali.align_to_mean_structure(structures, alignment,
                                                 start_structure = structures[0])
        
        mean_structure = mstrus[-1]
        
        # get alignments to mean structure
        rot, tr, perm, inv, dh = ali.get_RTPI(structures, mean_structure, alignment)
        
        for cc, spectra in enumerate(spectras):
            spectra.transform(rot=rot[cc], tr=tr[cc], perm=perm[cc],
                              inv=inv[cc], atype=alignment["type"])
        
        return
    
    def get_mean_structure(self, label):
        
        # check alignment was done
        assert self._aligned, "You might want to align the structures before..."
            
        # get spectras and structures
        spectras   = self.__get_atomic_label(label)
        structures = [sp.structure for sp in spectras]

        positions = [cc.get_positions() for cc in structures]

        mean_structure = structures[0].copy()
        mean_structure.set_positions(np.mean(positions, axis=0))

        return mean_structure
    
 
    # # generate a set of IAOS basis 
    # def generate_iaos_basis(self, minao="minao"):
        
    #     iaos_list    = []
    #     ref_spectras = []
    #     for lab in set(self.labels):
            
    #         if lab == -1: # ignore noise
    #             continue
            
    #         # get indexes where labels
    #         tidx = np.where(self.labels == lab)[0]
            
    #         # get them clusters
    #         clusters = self.comp.get_clusters_with_label(lab)
            
    #         # get distance matrix
    #         mean_structure = self.get_mean_structure(lab)
    #         dists = met.get_simple_distances(mean_structure, clusters)
            
    #         # find actual structure closest to mean structure and spectra
    #         ref_idx = tidx[np.argmin(dists)]
    #         spectra = self.spectras[ref_idx]
            
    #         # make the iaos
    #         iaos = self.make_iaos(spectra.mol, spectra._mo_coeff, spectra._mo_occ, minao)
    #         iaos_list.append(iaos)
    #         ref_spectras.append(spectra)

    #     return np.array(iaos_list), copy.deepcopy(ref_spectras)
    
    
    def get_feature_vector(self, label=None):
        
        if label is None:
            sp_list = self.spectras
        else:
            assert label in self.labels and label != -1, "Invalid label provided"
            sp_list = self.__get_atomic_label(label)
        
        energies   = []
        amplitudes = []
        overlaps   = []
        # CMOs       = []
        for spectra in sp_list:
                  
            # calculate overlap
            ovlp = spectra.get_CMO_orth_proj()
            
            # append data
            overlaps.append(ovlp**2)
            energies.append(spectra.energies)
            amplitudes.append(spectra.amplitude)

        return np.array(overlaps), np.array(energies), np.array(amplitudes)
    
    # make class iterable
    def __getitem__(self, index):
        if isinstance(index, (int, np.integer)):  # Include np.integer to handle NumPy integers
            return self.spectras[index]
        elif isinstance(index, slice):
            subset_spectras = self.spectras[index]
            return Spectras(subset_spectras)
        elif isinstance(index, (list, np.ndarray)):
            subset_spectras = [self.spectras[i] for i in index]
            return Spectras(subset_spectras)
        else:
            raise TypeError("Invalid index type")
    
    def __iter__(self):
        return iter(self.spectras)
    
    def __len__(self):
        return len(self.spectras)
    
    def __get_atomic_label(self, label):
        return [sp for sp in self if sp.label == label]
    
    def _prepare_for_save(self):
        
        data = self.__dict__.copy()
        
        spectras = [sp._prepare_for_save() for sp in self.spectras]
        
        data["spectras"] = spectras
        
        return data
    
    def save(self, filename="spectras.pkl"):
        """Saves the object to a file."""
        
        data = self._prepare_for_save()
      
        # Save to file
        with open(filename, 'wb') as fout:
            dill.dump(data, fout)
            
        return
    
    # use it to return a copy of the spectra collection object
    def copy(self):
        data = self._prepare_for_save()
        return Spectras(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 27 14:42:44 2023

@author: roncofaber
"""

import os
from pymbxas import __qcauxdir__, __qcdir__
#%%

def get_qchem_version_from_output(output):
    for line in output.splitlines():
        if line.strip().startswith("Q-Chem"):
            version = float(line.strip().split()[1])
            return version

def set_qchem_environment(scratchdir=None):

    # scratch in current directory if not specified
    if scratchdir is None:
        scratchdir = os.getcwd()

    # Clearixs qchem env. variables
    os.environ["QC"]         =  __qcdir__
    os.environ["QCAUX"]      = __qcauxdir__
    os.environ["QCTHREADS"]  = '1'
    os.environ["QCPARALLEL"] = 'True'
    os.environ["QCSCRATCH"]  = scratchdir

    # Path variables
    path_string = "{}/bin:{}/bin/perl".format(__qcdir__, __qcdir__)
    if path_string not in os.environ["PATH"]:
        os.environ["PATH"] += "{}/bin:{}/bin/perl".format(__qcdir__, __qcdir__)

    return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  1 18:13:29 2023

@author: roncoroni
"""

import os
import dill
import time
import logging

# good ol' numpy
import numpy as np

# self module utilities
import pymbxas
from pymbxas.calculators.excitation import Excitation
import pymbxas.utils.check_keywords as check
from pymbxas.utils.auxiliary import as_list
from pymbxas.utils.indexing import atoms_to_indexes
from pymbxas.io.data import pyscf_data
from pymbxas.io.logger import configure_logger
from pymbxas.build.structure import ase_to_mole
from pymbxas.build.input_pyscf import make_pyscf_calculator
from pymbxas.utils.orbitals import find_1s_orbitals_pyscf
from pymbxas.utils.boys import do_localization_pyscf
from pymbxas.mbxas.broaden import get_mbxas_spectra
from pymbxas.io.cleanup import remove_tmp_files
from pymbxas.io.write import write_data_to_fchk

# pyscf stuff
from pyscf.scf.addons import mom_occ
from pyscf import lo

#%%

class PySCF_mbxas():
    
    def __init__(self,
                 structure    = None,
                 charge       = 0,
                 spin         = 0,
                 xc           = "b3lyp",
                 basis        = "def2-svpd",
                 pbc          = False,
                 solvent      = None,
                 calc_type    = "UKS",
                 do_xch       = True,
                 pkl_file     = None,
                 target_dir   = None,
                 verbose      = 4,
                 print_fchk   = True,
                 print_output = False,
                 save         = True,  # save object as pkl file
                 save_chk     = False, # save calculation as chkfile
                 save_name    = "pyscf_obj.pkl", # name of saved file
                 save_path    = None, # path of saved object
                 loc_type     = "ibo",
                 gpu          = False,
                 ):
        
        # restart object from a pkl file and terminate
        if pkl_file is not None:
            self._restart_from_pickle(pkl_file)
            return
        
        # store directories and path
        self._cdir = os.getcwd() # current directory
        self._tdir = os.getcwd() if target_dir is None \
            else os.path.abspath(target_dir) # target directory

        if not os.path.exists(self._tdir):
            os.makedirs(self._tdir)


        if pkl_file is None:
            self._initialize_from_scratch(structure, charge, spin,
                                          xc, basis, pbc, solvent, calc_type,
                                          do_xch, verbose, print_fchk,
                                          print_output, save, loc_type,
                                          save_name, save_path, save_chk, gpu)
        
        return

    def _initialize_from_scratch(self, structure, charge, spin,
                                  xc, basis, pbc, solvent, calc_type,
                                  do_xch, verbose, print_fchk, print_output,
                                  save, loc_type, save_name, save_path,
                                  save_chk, gpu):

        # output, verbose and printing
        self._output_settings = {
            "print_fchk"   : print_fchk,
            "print_output" : print_output,
            "verbose"      : verbose,
            "save"         : save,
            "save_chk"     : save_chk,
            "save_path"    : save_path,
            "save_name"    : save_name,
            "is_gpu"       : gpu,
            }
        
        # logger
        configure_logger(verbose)
        self.logger = logging.getLogger(__name__)  # Logger tied to this class

        # check (#TODO in future allow to restart from a GS calculation)
        self._ran_GS   = False
        self._used_loc = False # becomes true if localization was needed

        # store calculation details
        self.structure = structure
        
        self._parameters = {
            "charge"   : charge,
            "spin"     : spin,
            "xc"       : xc,
            "basis"    : basis,
            "solvent"  : solvent,
            "pbc"      : check.check_pbc(pbc, structure),
            "loc"      : loc_type,
            "xch"      : do_xch,
            "calc_type": calc_type,
            # ... add more parameters as needed
            }
        
        # initialize empty stuff
        self.output       = {}
        self.data         = {}
        self.excitations  = []
        self.excited_idxs = []
              
        return
    
    # run all pymbxas from scratch
    def kernel(self, excitation):
        
        
        header = """----------------------------------|
           |                                  |
           |>>>>>>   Starting PyMBXAS   <<<<<<|
           |                                  |
           |       ver {:>7} | {:<12} |
           |----------------------------------|
        """.format(pymbxas.__version__, pymbxas.__date__)
        
        self.logger.info(header)
        
        # change directory
        os.chdir(self._tdir)
        
        # run ground state
        self.run_ground_state()
        
        # save object if needed
        if self.oset["save"]:
            self.save_object()
        
        # run all specified excitations
        self.excite(excitation)
        
        # save object if needed
        if self.oset["save"]:
            self.save_object()
            
            self.logger.info("Saved everything as {}".format(self.oset["save_name"]))
            
        # go back where we were
        os.chdir(self._cdir)
        
        self.logger.info("PyMBXAS finished successfully!")
        
        # clean up mess
        remove_tmp_files(self._tdir)
        
        return

    # excite an atom or a list of atoms
    def excite(self, ato_idxs, channel=1):
        
        if not self._ran_GS:
            self.logger.error("Please run a GS calculation first.")
            return
            
        
        # convert into atom indexes
        to_excite = atoms_to_indexes(self.structure, ato_idxs)
        
        # iterate over the indexes 
        for ato_idx in to_excite:
            self._single_excite(ato_idx, channel)
            
            # save object if needed
            if self.oset["save"]:
                self.save_object()
        
        return
    
    # perform a single excitation 
    def _single_excite(self, ato_idx, channel):
         
        if ato_idx in self.excited_idxs:
            return 
            
        excitation = Excitation(self.structure, self.gs_data, ato_idx,
                                self.parameters, channel, self.df_obj,
                                self.oset, self.logger)
        
        self.excited_idxs.append(ato_idx)
        self.excitations.append(excitation)
            
        return
        

    # run the GS calculation
    def run_ground_state(self, force=False):
        
        # check if GS was already performed, if so: skip
        if self._ran_GS and not force:
            self.logger.warn("GS already ran with this configuration. Skipping.")
            return
        self.logger.info("Started a new GS calculation")
        
        start_time  = time.time()

        # get system settings
        charge    = self.parameters["charge"]
        spin      = self.parameters["spin"]
        basis     = self.parameters["basis"]
        xc        = self.parameters["xc"]
        pbc       = self.parameters["pbc"]
        solvent   = self.parameters["solvent"]
        calc_type = self.parameters["calc_type"]
        
        # generate molecule
        gs_mol = ase_to_mole(self.structure, charge, spin, basis=basis, pbc=pbc,
                             verbose=self.oset["verbose"],
                             print_output=self.oset["print_output"],
                             is_gpu=self.oset["is_gpu"])
        
        # generate KS calculator
        gs_calc = make_pyscf_calculator(gs_mol, xc, pbc=pbc, solvent=solvent,
                                        calc_type=calc_type, dens_fit=None,
                                        calc_name="gs", save=self.oset["save_chk"],
                                        gpu=self.oset["is_gpu"])

        # run SCF #TODO: check how to change convergence parameters
        gs_calc.kernel()
        
        # store input/output
        self.output  = gs_calc.stdout.log.getvalue()
        self.gs_data = pyscf_data(gs_calc)
        self.mol     = gs_mol
    
        # store density object
        if pbc:
            # generate density fitter
            self.df_obj = gs_calc.with_df
        else:
            self.df_obj = None

        # check if localization is needed and run it
        self._run_localization(self.gs_data, self.parameters["loc"])
        
        # write output fchk files if using mokit
        if self.oset["print_fchk"]:
            self._print_fchk_files()

        # mark that GS has been run
        self._ran_GS = True
        
        self.logger.info(u"GS finished in {:.1f} s.\n".format(time.time() - start_time))
        return
    
    # run localization procedure
    def _run_localization(self, dft_calc, loc_type):
        
        # define list of relevant atoms
        ato_idxs = [cc for cc, ato in enumerate(self.structure) if ato.symbol != "H"]
        
        # check for degenerate delocalized orbitals and if necessary, do Boys
        s1_orbitals = []
        for ii in [0,1]:
            s1orb = find_1s_orbitals_pyscf(dft_calc.mol, dft_calc.mo_coeff[ii],
                                         dft_calc.mo_energy[ii],
                                         dft_calc.mo_occ[ii], as_list(ato_idxs),
                                         check_deg=True)
            s1_orbitals.append(s1orb)
        
        # only one s orb, no loc needed
        if len(s1_orbitals[1]) <= 1:
            return dft_calc.mo_coeff, False
        
        # localize up to highest degenerate orbital #TEST
        if loc_type.endswith("m"):
            s1_orbitals = [list(range(np.max(orb))) for orb in s1_orbitals]
            
        
        mo_loc = do_localization_pyscf(dft_calc, s1_orbitals, loc_type)
        
        self.logger.info("{} localization : {}".format(loc_type.upper(), s1_orbitals[1]))
        
        self._used_loc = True
        
        # update MO coeff if localization was used
        if self._used_loc:
            self.gs_data.mo_coeff_del = self.gs_data.mo_coeff.copy()
            self.gs_data.mo_coeff     = mo_loc
        
        return
    

    def _print_fchk_files(self):
        
    # write MOsmo_occ
        if self._used_loc:
        
            write_data_to_fchk(self.mol,
                              self.data,
                               oname    = self._tdir + "/output_gs_del.fchk",
                               mo_coeff = self.data.mo_coeff_del
                               )
            
            write_data_to_fchk(self.mol,
                               self.data,
                               oname    = self._tdir + "/output_gs_loc.fchk",
                               mo_coeff = self.data.mo_coeff
                               )
        
        else:
            write_data_to_fchk(self.mol,
                               self.data,
                               oname    = self._tdir + "/output_gs.fchk",
                               mo_coeff = self.data.mo_coeff
                               )
        
        # write LIVVOs
        mo_vvo = np.zeros(self.data.mo_coeff.shape)
        for channel in [0,1]:
            
            shape = self.data.mo_livvo.shape[1]
            
            mo_vvo[channel][:,:shape] = self.data.mo_livvo
        
        write_data_to_fchk(self.mol,
                           self.data,
                           oname    = self._tdir + "/livvos.fchk",
                           mo_coeff = self.data.mo_livvo
                           )
            
        return
    

    def get_mbxas_spectra(self, ato_idx, axis=None, sigma=0.02, npoints=3001, tol=0.01,
                          erange=None):
        
        ato_idxs = atoms_to_indexes(self.structure, ato_idx)
        
        energies    = []
        intensities = []
        for exc in self.excitations:
            if exc.ato_idx not in ato_idxs:
                continue
            energies.append(exc.mbxas["energies"])
            intensities.append(exc.mbxas["absorption"])
            
        energies = np.concatenate(energies)
        intensities = np.concatenate(intensities, axis=1)
        
        erange, spectras = get_mbxas_spectra(energies, intensities,
                                              sigma=sigma, npoints=npoints,
                                              tol=tol, erange=erange)

        if axis is None:
            spectras = np.mean(spectras, axis=0)
        else:
            spectras = spectras[axis]

        return erange, spectras

    # save object as pkl file
    def save_object(self, oname=None, save_path=None):
        
        if oname is None:
            oname = self.oset["save_name"]

        if save_path is None:
            path = self._tdir
        else:
            path = save_path

        if oname.endswith(".pkl"):
            oname =  path + "/" + oname
        else:
            oname = path + "/" + oname.split(".")[-1] + ".pkl"

        with open(oname, 'wb') as fout:
            dill.dump(self, fout)

        return
    
    # restart object from pkl file previously saved
    def _restart_from_pickle(self, pkl_file):

        # open previously generated gpw file
        with open(pkl_file, "rb") as fin:
            restart = dill.load(fin)

        self.__dict__ = restart.__dict__.copy()

        return
    
    @property
    def oset(self):
        return self._output_settings.copy()
    
    @property
    def parameters(self):
        return self._parameters.copy()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 27 11:19:06 2024

@author: roncoroni
"""

import numpy as np
from pyscf import data

#%%

def get_center_of_mass(mol):
    """Calculates the center of mass of a molecule.

    Args:
        coordinates (np.ndarray): Array of atomic coordinates (shape: (N, 3))
        masses (np.ndarray): Array of atomic masses (shape: (N))

    Returns:
        np.ndarray: Center of mass coordinates (shape: (3,))
    """
    
    coordinates = mol.atom_coords()
    masses = np.array([data.elements.COMMON_ISOTOPE_MASSES[m] #* data.nist.AMU2AU
                       for m in mol.atom_charges()])
    
    return np.average(coordinates, axis=0, weights=masses)


def get_inertia_tensor(mol, vectors=False):
    """Calculates the inertia tensor of a molecule.
    """
    
    Itensor = mol.inertia_moment()
    
    # evals, evecs = np.linalg.eigh(Itensor)
    evals, evecs = np.linalg.eig(Itensor)
    
    if vectors:
        return evals, evecs.transpose()
    else:
        return evals


def get_angular_momentum(mol, velocities):
    """Calculates the angular momentum of a molecule.

    Args:
        coordinates (np.ndarray): Array of atomic coordinates (shape: (N, 3))
        velocities (np.ndarray): Array of atomic velocities (shape: (N, 3))
        com (np.ndarray): Center of mass coordinates (shape: (3,))
        masses (np.ndarray): Array of atomic masses (shape: (N))

    Returns:
        np.ndarray: Angular momentum vector (shape: (3,))
    """
    
    coordinates = mol.atom_coords()
    masses = np.array([data.elements.COMMON_ISOTOPE_MASSES[m] #* data.nist.AMU2AU
                       for m in mol.atom_charges()])
    com = get_center_of_mass(mol)
    
    coordinates -= com  # Center coordinates around COM
    
    return np.sum(np.cross(coordinates, masses[:, np.newaxis] * velocities), axis=0)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 15:15:11 2023

@author: roncoroni
"""

import numpy as np

# check if cupy is available
try:
    import cupy as cp
except ImportError:
    cp = None 

import copy

#%%

# class to read and store data from a pyscf calculation
class pyscf_data():
    """
    This class provides a convenient wrapper for storing and manipulating 
    data extracted from a PySCF calculation. It supports conversion between 
    NumPy (CPU) and CuPy (GPU) array formats for potential acceleration.
    """
    def __init__(self, calculator):
        """
        Initializes the pyscf_data object.
        
        Args:
            calculator (pyscf.gto.Mole, pyscf.scf.HF, etc.): A PySCF calculator 
                object that has already been run. If None, an empty object is created. 
        """
        
        self.mol       = calculator.mol
        self.mo_coeff  = calculator.mo_coeff
        self.mo_occ    = calculator.mo_occ
        self.mo_energy = calculator.mo_energy
        self.e_tot     = calculator.e_tot
        self.nelec     = calculator.nelec
        
        # Convert arrays in-place to ensure np.array
        if cp is not None:
            for attr_name in vars(self):
                attr_value = getattr(self, attr_name)
                if isinstance(attr_value, cp.ndarray):
                    setattr(self, attr_name, attr_value.get())
                
        self._is_gpu = False
        
        return
    
    def copy(self):
        return copy.deepcopy(self)
    
    def to_cpu(self):
        """Converts all internal arrays to NumPy format"""
        
        result = self.copy()  # Create a placeholder instance
        
        if not result._is_gpu:
            return result

        # Iterate through attributes and convert CuPy arrays
        for attr_name in vars(self):
            attr_value = getattr(self, attr_name)
            if isinstance(attr_value, cp.ndarray):
                result.__setattr__(attr_name, attr_value.get())  
            else:
                result.__setattr__(attr_name, attr_value)
                
        result._is_gpu = False

        return result
    
    def to_gpu(self):
        """Converts all internal arrays to CuPy format"""
        
        result = self.copy()  # Create a placeholder instance
        
        if result._is_gpu:
            return result
        
        # Iterate through attributes and convert NumPy arrays
        for attr_name in vars(self):
            attr_value = getattr(self, attr_name)
            if isinstance(attr_value, np.ndarray):
                result.__setattr__(attr_name, cp.asarray(attr_value))  
            else:
                result.__setattr__(attr_name, attr_value)
                
        result._is_gpu = True
    
        return result


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 23 10:25:44 2023

@author: roncofaber
"""

import os

#%%

def submit_qchem_job(run_path, partition, account, procs, time,
                     output="qchem.output"):
    
        file = run_path + "/qchemjob.sh"
        
        if os.path.isfile(file) == True:
            os.remove(file)
        
        with open(file, "a") as fout:
            fout.write("#!/bin/bash\n\n")
            fout.write("#SBATCH --job-name=mbxas.qchem\n")
            fout.write("#SBATCH --partition="+partition+"\n")
            fout.write("#SBATCH --time="+time+"\n")
            fout.write("#SBATCH -N 1 \n")
            fout.write("#SBATCH --account="+account+"\n")
            fout.write("#SBATCH --chdir="+run_path+"\n")
            
            if partition.startswith("lr"):
                fout.write("#SBATCH --qos=lr_normal\n")
            
            fout.write("export  QC=/clusterfs/etna/pscratch/subhayan/QCHEM_CODE/qchem-trunk-34979\n")
            
            fout.write("export QCSCRATCH="+run_path+"\n")
            fout.write("export SCRATCH="+run_path+"\n")
            
            fout.write("export QCTHREADS=1\n")
            fout.write("export QCPARALLEL=True\n")
            fout.write("export QCAUX=/clusterfs/etna/pscratch/subhayan/QCHEM_CODE/qcaux-trunk\n")
            fout.write("export PATH=$PATH:$QC/bin:$QC/bin/perl\n")
            
            fout.write("module purge\n")
            
            for module in ["gcc/6.3.0", "fftw/3.3.6-gcc",
                           "openmpi/3.0.1-gcc", "mkl/2016.4.072"]:
                fout.write("module load {}\n".format(module))
                
            # fout.write("module load openmpi\n")
            fout.write("cd {}\n".format(run_path))
            
            fout.write("qchem -nt {} -save qchem.input > input.qchem.out {} saved_files".format(
                procs, output))
        
        print("Job submitted in: {}".format(run_path))
        
        os.chdir(run_path)
        os.system("sbatch {}/qchemjob.sh".format(run_path))
            
        return
    
    
def submit_mbxas_job(run_path, mbxas_dir):
    
    os.chdir(run_path)
    with open("{}/clearixsjob.sh".format(run_path), "w") as fout:
        fout.write("#!/bin/bash\n\n")
        fout.write("python {}/RIXS_main.py {}/INP_FILE\n".format(
            mbxas_dir, run_path))

    os.system("bash {}/clearixsjob.sh".format(run_path))
    
    return
             


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 12 15:36:48 2024

@author: roncofaber
"""

# ase stuff
import ase
from ase import units
import ase.visualize
from ase.calculators.singlepoint import SinglePointCalculator

# system stuff
import os
# this repo
from pymbxas.build.structure import mole_to_ase

#%%

au_2_fs = 0.02418884254

# class to add to a AIMD callback to print traj every n steps as ase atoms xyz
class AIMDTrajWriter():
    
    def __init__(self,
                 oname  = "aimd_trajectory.xyz", # name output traj
                 nstep  = 1,    # frequency at which traj is written
                 data_output  = None, # also write output
                 append       = False # append to existing traj
                 ):
        
        self.nstep = nstep
        self.oname = oname
        
        if os.path.exists(oname) and not append:
            os.remove(oname)
            
        self._setup_data_output(data_output)
        
        return
    
    def _setup_data_output(self, data_output):
        
        # avoid opening data_output file twice
        if type(data_output) is str:
            if os.path.isfile(data_output):
                print('overwrite data output file: %s' %
                      data_output)
            else:
                print('data output file: %s' % data_output)

            if data_output == '/dev/null':
                self.data_output = open(os.devnull, 'w')

            else:
                self.data_output = open(data_output, 'w')
                self.data_output.write(
                    'time          Epot                 Ekin                 '
                    'Etot                 T\n'
                )
        
        return
    
    def _write_data(self, data):
        '''Writes out the potential, kinetic, and total energy, temperature to the
        self.data_output stream. '''

        output = '%8.2f  %.12E  %.12E  %.12E %3.4f' % (data.time,
                                                       data.epot,
                                                       data.ekin,
                                                       data.ekin + data.epot,
                                                       data.temperature())

        # We follow OM of writing all the states at the end of the line
        if getattr(data.scanner.base, 'e_states', None) is not None:
            if len(data.scanner.base.e_states) > 1:
                for e in data.scanner.base.e_states:
                    output += '  %.12E' % e

        self.data_output.write(output + '\n')

        # If we don't flush, there is a possibility of losing data
        self.data_output.flush()
        
        return
    
    def __call__(self, aimd):
        
        data = aimd["self"]
        
        # if not multiple, do nothing unless is last frame
        if data._step % self.nstep != 0:
            if data._step == data.steps:
                pass
            else:
                return
        
        # convert mol to ase
        atoms = mole_to_ase(data.mol)
    
        atoms.info["epot"]  = data.epot
        atoms.info["ekin"]  = data.ekin
        atoms.info["veloc"] = data.veloc
        atoms.info["etot"]  = data.epot + data.ekin
        atoms.info["time"]  = data.time
        
        # set velocity
        atoms.set_velocities(data.veloc*units.Bohr/(au_2_fs*units.fs))
        
        # assign energy as single point calc #TODO expand with forces and such
        calc = SinglePointCalculator(atoms, energy=data.epot*units.Ha)
        atoms.set_calculator(calc)
        
        # write traj
        ase.io.write(self.oname, atoms, append=True)
        
        if self.data_output is not None:
            self._write_data(data)
        return
    
# class to callback a geometry optimization
class OptimizerTraj():
    
    def __init__(self, write=True, oname="optimization.xyz", append=False):
        
        self._traj   = []
        self.oname  = oname
        self._write = write
        
        if write and os.path.exists(oname) and not append:
            os.remove(oname)
            
        return
    
    # function that is called by the optimizer
    def __call__(self, opt):

        atoms = mole_to_ase(opt["mol"])
        atoms.info["etot"] = opt["energy"]
        
        # assign energy as single point calc #TODO expand with forces and such
        calc = SinglePointCalculator(atoms, energy=opt["energy"]*units.Ha)
        atoms.set_calculator(calc)
        
        # update optimizer internal data
        self.append(atoms)
        self.set_mol(opt["mol"], atoms)
        
        # write output if specified
        if self._write:
            ase.io.write(self.oname, atoms, append=True)
        
        return
    
    def set_mol(self, mol, atoms=None):
        self._mol   = mol
        
        if atoms is None:
            atoms = mole_to_ase(mol)
        else:
            atoms = atoms
        self._atoms = atoms
        return
    
    def append(self, element):
        self._traj.append(element)
        return
    
    def view(self):
        ase.visualize.view(self)
        return
    
    # make class iterable
    def __getitem__(self, index):
        return self._traj[index]
    
    def __iter__(self):
        return iter(self._traj)
    
    # get elements
    @property
    def mol(self):
        return self._mol
    @property
    def traj(self):
        return self._traj
    @property
    def atoms(self):
        return self._atoms


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 18:12:42 2023

@author: roncoroni
"""

import os
import glob

#%%

def remove_tmp_files(current_directory):
    
    file_list = glob.glob(current_directory + "/tmp*")
    
    for file in file_list:
        if os.path.isfile(file):
            os.remove(file)
            
    return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 22 12:01:06 2023

@author: roncofaber
"""

import os
# import pymbxas.build.input as inp
import numpy as np

# MOKIT stuff
try:
    import mokit
    from mokit.lib.py2fch import py2fch
    from mokit.lib.py2fch_direct import mol2fch
    is_mokit = True
except:
    try:
        from mokit.lib.py2fch import py2fch
        from mokit.lib.py2fch_direct import mol2fch
        is_mokit = True
    except:
        is_mokit = False
    
# from mokit.lib.py2fch import py2fch
# from mokit.lib.py2fch_direct import mol2fch

#%%

def write_data_to_fchk(mol, mo_coeff=None, mo_occ=None, density=False,
                       mo_energy=None, oname="tmp.fchk",  center=False):
    
    # if not is_mokit:
    #     print("No MOKIT")
    #     return
    
    directory = os.path.dirname(oname)
    if directory and not os.path.exists(directory):
        os.makedirs(directory)
        
    mol = mol.copy()
    
    if center:
        origin = np.mean([ii[1] for ii in mol.atom], axis=0)
        
        new_atom = mol.atom.copy()
        for cc, atom in enumerate(new_atom):
            new_atom[cc][1] = np.array(atom[1]) - origin
        
        mol.set(atom=new_atom)
        mol.build()
    
    if mo_coeff is None:
        mo_coeff = mol.intor_symmetric('int1e_ovlp')
    
    if len(mo_coeff) == 2:
        nbasis, norbit = mo_coeff[0].shape
    else:
        nbasis, norbit = mo_coeff.shape
        mo_coeff = np.array([mo_coeff, mo_coeff])
    
    if mo_energy is None:
        pass
    elif not len(mo_energy) == 2:
        mo_energy = np.array([mo_energy, mo_energy])
        
    # update mo_occ just to make sure
    if mo_occ is not None:
        nelec = mo_occ.sum(axis=1, dtype=int)
    else:
        nelec = list(mol.nelec)
    
    for cc, ne in enumerate(nelec):
        
        if ne > mo_coeff[cc].shape[1]:
            nelec[cc] = mo_coeff[cc].shape[1]
        
            # nelec[cc] = np.minimum(ne, mo_coeff.shape[cc+1])
    
    mol.nelec = nelec
        
    # shape doesn't match, means that probably the energy is useless
    if mo_energy is None:
        mo_energy = np.zeros((2, norbit))
    if mo_energy.shape[1] != norbit:
        mo_energy = np.zeros((2, norbit))
        
    if norbit > mo_coeff[0].shape[1]:
        print("Cut orbital number cause higher than allowed by IQmol")
        
        norbit_max = mo_coeff[0].shape[1]
    else:
        norbit_max = norbit
    
    # actual write
    mol2fch(mol, oname , True, mo_coeff[:,:,:norbit_max])
    for cc, spin in enumerate(["a", "b"]):
        py2fch(oname, nbasis, norbit_max, mo_coeff[cc][:,:norbit_max], spin,
               mo_energy[cc][:norbit_max], False, density)
    
    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 15 11:23:47 2023

@author: roncoroni
"""

import subprocess
import psutil
import numpy as np
import collections.abc

#%%
# get available memory either for CPU or GPU
def get_available_memory(is_gpu=False):
    if is_gpu:
        try:
            command = "nvidia-smi --query-gpu=memory.free --format=csv"
            memory_free_info = subprocess.check_output(command.split()).decode('ascii').split('\n')[:-1][1:]
            memory_free_values = [int(x.split()[0]) for x in memory_free_info]
            return memory_free_values[0]  # Return all GPU memory values
        except FileNotFoundError:
            print("nvidia-smi is not installed or not found in PATH.")
            return None
        except subprocess.CalledProcessError as e:
            print(f"Error occurred while executing nvidia-smi: {e}")
            return None
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            return None
    else:
        try:
            return int(psutil.virtual_memory().available / 1e6)
        except Exception as e:
            print(f"An unexpected error occurred while retrieving system memory: {e}")
            return None

    
    
# return copy of input as list if not one
def as_list(inp):
    if inp is None:
        return None
    elif isinstance(inp, int) or isinstance(inp, np.int64):
        return [inp]
    elif isinstance(inp, collections.abc.Iterable) and not isinstance(inp, str): 
        # Handles lists, tuples, NumPy arrays, etc. (Excludes strings)
        return list(inp)  
    else:
        raise TypeError(f"Cannot convert type {type(inp)} to list")
    

def s2i(string):
    if string.lower() == "beta":
        return 1
    elif string.lower() == "alpha":
        return 0
    else:
        raise TypeError
     
# see the docstring
def standardCount(label_arr, labels = None):
    '''
    Most Common Number of occurances of each label in each row of label_arr

    Parameters
    ----------
    labels : array (n_labels)
        List of target labels. If None assumes all unique elements of label_arr
    label_arr : array (n_rows, n_col)
        reference array to find most common elements.

    Returns
    -------
    List of most common frequency of each label (n_labels).

    '''
    labels = as_list(labels)
    
    label_arr = np.array(label_arr)
    
    if labels is None:
        labels = set(label_arr.reshape(-1))
        labels.discard(-1)
    ## How many occurances of a label in each structure
    label_count = np.array([np.sum(label_arr == label, axis= 1) for label in labels]).T
    
    ## The standard number of occurances of each cluster
    unique, counts = np.unique(label_count, axis= 0, return_counts= True)
    standard_count = unique[counts.argmax()]
    
    return np.squeeze(standard_count)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul  7 12:28:13 2023

@author: roncofaber
"""

import os
import dill
import time

# good ol' numpy
import numpy as np

# self module utilities
import pymbxas
import pymbxas.utils.check_keywords as check
from pymbxas.io.copy import copy_output_files
from pymbxas.build.input import make_qchem_input
from pymbxas.utils.orbitals import find_1s_orbitals, calculate_boys_overlap
from pymbxas.utils.environment import set_qchem_environment, get_qchem_version_from_output
from pymbxas.utils.basis import get_basis_set_info
import pymbxas.CleaRIXS.read_modules as rd
import pymbxas.CleaRIXS.rixs_modules as rx

# pyqchem stuff
from pyqchem import get_output_from_qchem
from pyqchem.file_io import write_to_fchk
from pyqchem.parsers.basic import basic_parser_qchem

#%%

class Qchem_mbxas():

    def __init__(self,
                 structure    = None,
                 charge       = None,
                 multiplicity = None,
                 pkl_file     = None,
                 qchem_params = None,
                 excitation   = None,
                 do_xch       = True,
                 target_dir   = None,
                 scratch_dir  = None,
                 print_fchk   = True,
                 run_calc     = True,
                 use_mpi      = False, # somewhat MPI is not working atm
                 use_boys     = True,  # use Boys localization or not
                 save         = True,  # save object as pkl file
                 save_all     = True, # save all fchk output in the pkl
                 save_name    = "mbxas_obj.pkl", # name of saved file
                 save_path    = None, # path of saved object
                 keep_scratch = False,
                 ):

        # restart object from a pkl file
        if pkl_file is not None:
            self.__restart_from_pickle(pkl_file)

        # set up internal variables
        self.__is_mpi = use_mpi
        self.__nprocs = os.cpu_count()
        self.__pid    = os.getpid()

        # store directories and path
        self.__cdir = os.getcwd() # current directory
        self.__tdir = os.getcwd() if target_dir is None \
            else os.path.abspath(target_dir) # target directory
        self.__sdir = self.__tdir if scratch_dir is None else scratch_dir # scratch directory
        self.__wdir = "{}/pyqchem_{}/".format(self.__sdir, self.__pid) # PyQCHEM work directory

        if not os.path.exists(self.__tdir):
            os.makedirs(self.__tdir)

        # initialize environment (set env variables)
        set_qchem_environment(self.__sdir)

        if pkl_file is None:
            self.__initialize_from_scratch(structure, charge, multiplicity,
                                           qchem_params, excitation, use_boys,
                                           do_xch, print_fchk, save, save_name,
                                           save_path, save_all, keep_scratch,
                                           run_calc)

        return

    def __initialize_from_scratch(self, structure, charge, multiplicity,
                                  qchem_params, excitation, use_boys, do_xch,
                                  print_fchk, save, save_name, save_path,
                                  save_all, keep_scratch, run_calc):

        # calculation details
        self.__use_boys   = use_boys
        self.__is_xch = do_xch

        # output, verbose and printing
        self.__print_fchk = print_fchk
        self.__save_all = save_all
        self.__keep_scratch = keep_scratch

        # check (#TODO in future allow to restart from a GS calculation)
        self.__ran_GS = False

        # store data
        self.structure    = structure
        self.charge       = charge
        self.multiplicity = multiplicity
        self.qchem_params = qchem_params
        self.excitation   = check.determine_excitation(excitation)

        # initialize empty stuff
        self.input = {}
        self.output = {}
        self.data   = {}

        # run MBXAS calculation
        if run_calc:
            self.run_all_calculations()

        if save:
            self.save_object(save_name, save_path)

        return

    # restart object from pkl file previously saved
    def __restart_from_pickle(self, pkl_file):

        # open previously generated gpw file
        with open(pkl_file, "rb") as fin:
            restart = dill.load(fin)

        self.__dict__ = restart.__dict__.copy()

        return

    # Function to run all calc (GS, FCH, XCH) in sequence
    def run_all_calculations(self):

        print(">>> Starting PyMBXAS <<<")

        start_time  = time.time()

        # run ground state
        print("> Ground state calculation: ", end = "", flush=True)
        gs_output, gs_data = self.run_ground_state()
        gs_time = time.time() - start_time
        print("finished in {:.2f} s.".format(gs_time))

        print("> Exciting orbital #{} - {} orbs.".format(self.excitation["eject"],
                                                    ["KS", "Boys"][self.__use_boys]))
        # run FCH
        print("> FCH calculation: ", end = "", flush=True)
        scf_guess = gs_data["localized_coefficients"] if self.__use_boys else gs_data["coefficients"]
        fch_output, fch_data  = self.run_fch(scf_guess) #TODO change only if Boys
        fch_time = time.time() - gs_time - start_time
        print("finished in {:.2f} s.".format(fch_time))

        tot_time = fch_time
        # only run XCH if there is input
        xch_time = None,
        if self.__is_xch:
            print("> XCH calculation: ", end = "", flush=True)
            xch_output, xch_data = self.run_xch(fch_data["coefficients"])
            xch_time = time.time() - fch_time - gs_time - start_time

            if xch_output is None: #failed
                print("failed! - ignoring XCH")

            else:
                print("finished in {:.2f} s.".format(xch_time))
            tot_time += xch_time

        print("> MBXAS calculation: ", end = "", flush=True)
        self.run_mbxas()
        mbxas_time = time.time() - tot_time - gs_time - start_time
        print("finished in {:.2f} s.".format(mbxas_time))

        self.timings = {
            "gs"  : gs_time,
            "fch" : fch_time,
            "xch" : xch_time,
            "mbxas" : mbxas_time
            }

        print(">>> PyMBXAS finished successfully! <<<\n")

        return

    # run the GS calculation
    def run_ground_state(self):

        # check if GS was already performed, if so: skip
        if self.__ran_GS:
            print("GS already ran with this configuration. Skipping.")
            return self.output["gs"], self.data["gs"]

        # get system settings
        structure    = self.structure
        charge       = self.charge
        multiplicity = self.multiplicity
        qchem_params = self.qchem_params

        # GS input
        gs_input = make_qchem_input(structure, charge, multiplicity,
                             qchem_params, "gs", occupation = None,
                             use_boys=self.__use_boys)

        # run calculation
        gs_output, gs_data = get_output_from_qchem(
            gs_input, processors = self.__nprocs, use_mpi = self.__is_mpi,
            return_electronic_structure = True, scratch = self.__sdir,
            delete_scratch = False)

        # parse output (#TODO in future we won't need to print the output so it can be done directly)
        gs_data.update(basic_parser_qchem(gs_output))

        # read basis set information and store them
        atom_coeffs, atom_labels, symbols, nbasis, indexing = get_basis_set_info(gs_data["basis"])
        self.basis = {
            "atom_coefficients" : atom_coeffs,
            "atom_labels"       : atom_labels,
            "symbols"           : symbols,
            "nbasis"            : nbasis,
            "indexing"          : indexing,
            }

        # old version, loc coeff need to be swapped and override qchem_order
        gs_data = self.__check_basis_indexing(gs_data, gs_output, self.basis["indexing"])

        # obtain number of electrons
        #TODO make a function that stores relevant output (but not too heavy stuff)
        self.n_alpha = gs_data["number_of_electrons"]["alpha"]
        self.n_beta  = gs_data["number_of_electrons"]["beta"]
        self.n_electrons = self.n_alpha + self.n_beta

        # update with correct number of electrons
        self.excitation["nelectrons"] = [self.n_alpha, self.n_beta]

        # store input/output
        self.input["gs"]  = gs_input
        self.output["gs"] = gs_output
        self.basis_overlap = gs_data["overlap"]
        if self.__save_all:
            self.data["gs"] = gs_data

        # do boys postprocessing to understand orbital occupations
        if self.__use_boys:
            self.boys_coeffs = gs_data["localized_coefficients"]

        # find 1s orbitals
        self.s_orbitals  = find_1s_orbitals(gs_data, use_localized=self.__use_boys)
        # overwrite occupation if it's not been specified
        if not self.excitation["eject"]:
            to_eject = self.s_orbitals[
                self.excitation["channel"]][self.excitation["ato_idx"]]
            self.excitation["eject"] = to_eject + 1

        # write output file
        #TODO change in the future to be more flexible
        with open(self.__tdir + "/qchem.output", "w") as fout:
            fout.write(gs_output)

        if self.__print_fchk:
            write_to_fchk(gs_data, self.__tdir + "/output_gs.fchk")

        # mark that GS has been run
        self.__ran_GS = True

        return gs_output, gs_data

    # run the FCH calculation
    def run_fch(self, scf_guess=None):

        assert self.__ran_GS, "Please run a GS calculation first."

        structure    = self.structure
        charge       = self.charge + 1 # +1 cause we kick out one lil electron
        multiplicity = abs(self.n_alpha - self.n_beta - 1) + 1
        qchem_params = self.qchem_params
        excitation   = self.excitation

        # FCH input
        fch_input = make_qchem_input(structure, charge, multiplicity,
                                     qchem_params, "fch",
                                     occupation = excitation,
                                     scf_guess  = scf_guess)

        # run calculation
        fch_output, fch_data = get_output_from_qchem(
            fch_input, processors = self.__nprocs, use_mpi = self.__is_mpi,
            return_electronic_structure = True, scratch = self.__sdir,
            delete_scratch = not (self.__is_xch or self.__keep_scratch))

        # parse output (#TODO in future we won't need to print the output so it can be done directly)
        fch_data.update(basic_parser_qchem(fch_output))

        # old version, loc coeff need to be swapped and override qchem_order
        fch_data = self.__check_basis_indexing(fch_data, fch_output, self.basis["indexing"])

        # store input/output
        self.input["fch"]  = fch_input
        self.output["fch"] = fch_output
        if self.__save_all:
            self.data["fch"] = fch_data

        # write input and output
        with open(self.__tdir + "/qchem.input", "w") as fout:
            fout.write(fch_input.get_txt())
        with open(self.__tdir + "/qchem.output", "a") as fout:
            fout.write(fch_output)
        # copy MOM files in relevant directory
        copy_output_files(self.__wdir, self.__tdir)

        if self.__print_fchk:
            write_to_fchk(fch_data, self.__tdir + "/output_fch.fchk")

        # calculate overlap if using BOYS
        if self.__use_boys:

            self.boys_overlap = calculate_boys_overlap(self.boys_coeffs,
                                                       fch_data["coefficients"],
                                                       self.basis_overlap)

            # overwrite files
            for channel, overlap in self.boys_overlap.items():
                np.savetxt(self.__tdir + "/{}_ovlp_gs_es.txt".format(
                    channel), overlap)

        return fch_output, fch_data

    # run the XCH calculation
    def run_xch(self, scf_guess=None):

        assert self.__ran_GS, "Please run a GS calculation first."

        structure = self.structure
        charge = self.charge
        multiplicity = self.multiplicity
        qchem_params = self.qchem_params

        # xch occupation is always the same
        channel = self.excitation["channel"]
        channel_idx = 0 if  channel == "alpha" else 1
        nelec = [self.n_alpha, self.n_beta][channel_idx]
        xch_occ = {
            "nelectrons" : [self.n_alpha, self.n_beta],
            "eject"      : nelec,
            "inject"     : nelec + 1, #nelec + 1
            "channel"    : channel,
            }

        # XCH input
        xch_input = make_qchem_input(structure, charge, multiplicity,
                                     qchem_params, "xch", occupation=xch_occ,
                                     scf_guess=scf_guess)

        # try to run calculation
        try:
            xch_output, xch_data = get_output_from_qchem(
                xch_input, processors = self.__nprocs, use_mpi = self.__is_mpi,
                return_electronic_structure = True, scratch = self.__sdir,
                delete_scratch = not self.__keep_scratch)

            self.__xch_failed = False

        except: #calculation failed

            self.__is_xch = False
            self.__xch_failed = True

            return None, None


        # parse output (#TODO in future we won't need to print the output so it can be done directly)
        xch_data.update(basic_parser_qchem(xch_output))

        # old version, loc coeff need to be swapped and override qchem_order
        xch_data = self.__check_basis_indexing(xch_data, xch_output, self.basis["indexing"])

        # store input/output
        self.input["xch"]  = xch_input
        self.output["xch"] = xch_output
        if self.__save_all:
            self.data["xch"] = xch_data

        if self.__print_fchk:
            write_to_fchk(xch_data, self.__tdir + "/output_xch.fchk")

        # generate AlignDir directory #TODO change for more flex
        try:
            os.mkdir(self.__tdir + "/AlignDir")
        except OSError:
            pass

        # write XCH output file
        with open(self.__tdir + "/AlignDir/align_calc.out", "w") as fout:
            fout.write(xch_output)

        return xch_output, xch_data

    def run_mbxas(self):

        # Hartree to eV
        Ha = 27.211407953

        # select excitation channel
        channel = self.excitation["channel"]

        # get excitation info
        nocc = self.data["fch"]["number_of_electrons"][channel]
        core_ind_gs = self.excitation["eject"]

        # read matrices
        full_mom_matrix, full_gs_matrix, full_ovlp_matrix = rd.read_full_matrices(
            nocc, core_ind_gs, path=self.__tdir, channel=channel)

        # read energy of unoccupied states #TODO check consistency
        ener = np.array(self.data["fch"]["mo_energies"][channel][nocc+1:])

        # align energies
        if self.__is_xch:
            ener += self.data["xch"]["scf_energy"] - self.data["gs"]["scf_energy"] - np.min(ener)
        # else:
            # ener += self.data["fch"]["scf_energy"] #TODO check this

        ener = ener*Ha # convert energy to eV

        # calculate overlap and store absorption
        xi = np.array(full_ovlp_matrix).T
        norb = min(xi.shape)

        absorption = []
        for ixyz in [0, 1, 2]:
            chb_xmat = np.array(full_mom_matrix)[ixyz, :]
            absorption.append(rx.AbsCalc(ixyz, xi, nocc, norb, chb_xmat))

        self.mbxas = {
            "absorptions" : absorption,
            "energies"    : ener
            }

        return

    @staticmethod
    def __check_basis_indexing(data, output, indices):

        # override ordering to be consistent (53.0 can be overwritten by Boys)
        data["coefficients"]["qchem_order"] = indices

        if "localized_coefficients" not in data:
            return data

        data["localized_coefficients"]["qchem_order"] = indices

        if get_qchem_version_from_output(output) < 6:

            reverse_indices = [list(indices).index(j) for j in range(len(indices))]

            for channel in ["alpha", "beta"]:
                data['localized_coefficients'][channel] = \
                    np.array(data['localized_coefficients'][channel])[:, reverse_indices].tolist()

            data['localized_coefficients']["qchem_order"] = indices

        return data

    # broaden spectrum to plot it
    @staticmethod
    def broadened_spectrum(x, energies, intensities, sigma):

        intensities = np.asarray(intensities)

        def gaussian_broadening(x, sigma):
            return np.exp(-x**2 / (2 * sigma**2)) / (sigma * np.sqrt(2 * np.pi))

        broadened_spec = np.zeros_like(x)
        for energy, intensity in zip(energies, intensities**2): #square intensity
            broadened_spec += intensity * gaussian_broadening(x - energy, sigma)

        return broadened_spec

    # function to get MBXAS spectra
    def get_mbxas_spectra(self, sigma=0.3, npoints=3001, tol=0.01):

        min_E = np.min(self.mbxas["energies"])
        max_E = np.max(self.mbxas["energies"])

        dE = max_E - min_E

        energy = np.linspace(min_E - tol*dE, max_E + tol*dE, npoints)

        spectra = self.broadened_spectrum(energy, self.mbxas["energies"],
                                          np.mean(self.mbxas["absorptions"], axis=0),
                                          sigma)

        return energy, spectra


    # save object as pkl file
    def save_object(self, oname=None, save_path=None):

        if oname is None:
            oname = self.savename

        if save_path is None:
            path = self.__tdir
        else:
            path = save_path

        if oname.endswith(".pkl"):
            oname =  path + "/" + oname
        else:
            oname = path + "/" + oname.split(".")[-1] + ".pkl"

        with open(oname, 'wb') as fout:
            dill.dump(self, fout)

        print("Saved everything as {}".format(oname))
        return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  8 17:05:13 2023

@author: roncoroni
"""

# pyscf stuff
from pyscf import scf, dft
import pyscf.pbc as pypbc

#%%

def make_density_fitter(mol, pbc, cderi=False):
    
    if pbc:
        df_obj = pypbc.df.DF(mol)
        
        # df_obj.auxbasis = pypbc.df.make_auxbasis(mol)
        
        if cderi is not None:
            df_obj._cderi_to_save = 'saved_cderi.h5'
    else:
        raise "Density fitting for non PBC systems has yet to be implemented!"
    
    df_obj.build()
    
    return df_obj

# Function to make a pyscf calculator that both work with PBC and not.
def make_pyscf_calculator(mol, xc=None, calc_type="UKS", pbc=False, solvent=None,
                          dens_fit=None, calc_name=None, save=False, gpu=False):
    
    # with PBC
    if pbc:
        calc = getattr(pypbc.scf, calc_type)(mol, xc=xc).density_fit()
        
        # density fit object is string
        if dens_fit is not None:
            # calc.with_df._cderi = 'saved_cderi.h5'
            calc.with_df = dens_fit
    
    # no PBC
    else:
        
        # generate calculator
        if "HF" in calc_type.upper():
            if xc is not None:
                print("HF method: XC keyword ignored.")
            calc = getattr(scf, calc_type)(mol)
        elif "KS" in calc_type.upper():
            if xc is None:
                xc = "LDA" #default to LDA
            calc = getattr(dft, calc_type)(mol, xc=xc)
        
    # add solvent treatment
    if solvent is not None:
        calc = calc.DDCOSMO()
        calc.with_solvent.eps = solvent

    # Use chkfile to store calculation (if you want)
    if isinstance(calc_name, str) and save:
       calc.chkfile = '{}.chkfile'.format(calc_name)
    
    # add GPU compatibility
    if gpu:
        try:
            calc = calc.to_gpu()
        except ImportError:
            raise RuntimeError("CuPy is required for GPU calculations. Please install it.") 
            
    return calc
